import os
import json
import asyncio
from pathlib import Path
from datetime import datetime, timedelta
from typing import Dict, Any, Optional, List, Union, Literal
from typing_extensions import Self

import ulid
from pydantic import BaseModel, ConfigDict, RootModel, Field, field_validator, computed_field, model_validator, model_serializer, field_serializer

from aicore.logger import _logger
from aicore.const import DEFAULT_OBSERVABILITY_DIR, DEFAULT_OBSERVABILITY_FILE, DEFAULT_ENCODING

class LlmOperationRecord(BaseModel):
    """Data model for storing information about a single LLM operation."""
    session_id: Optional[str] = ""
    workspace: Optional[str] = ""
    agent_id: Optional[str] = ""
    action_id: Optional[str] = ""
    operation_id: str = Field(default_factory=ulid.ulid)
    timestamp: Optional[str] = ""
    operation_type: Literal["completion", "acompletion", "acompletion.tool_call"]
    provider: str
    input_tokens: Optional[int] = 0
    output_tokens: Optional[int] = 0
    cached_tokens: Optional[int] = 0
    cost: Optional[float] = 0
    latency_ms: float
    error_message: Optional[str] = ""
    extras: Union[Dict[str, Any], str] = ""
    completion_args: Union[Dict[str, Any], str]
    response: Optional[Union[str, Dict, List]] = ""

    model_config = ConfigDict(
        arbitrary_types_allowed = True
    )

    @field_validator(*["session_id", "workspace", "agent_id", "action_id", "timestamp", "error_message", "response"])
    @classmethod
    def ensure_non_nulls(cls, value: Optional[str] = None) -> str:
        if value is None:
            return ""
        return value

    @field_validator("response")
    @classmethod
    def json_dumps_response(cls, response: Union[None, str, Dict[str, str]]) -> Optional[str]:
        if isinstance(response, (str, type(None))):
            return response
        elif isinstance(response, (dict, list)):
            return json.dumps(response, indent=4)
        else:
            raise TypeError("response param must be [str] or [json serializable obj]")

    @field_validator(*["completion_args", "extras"])
    @classmethod
    def json_laods_response(cls, args: Union[str, Dict[str, Any]]) -> Dict[str, Any]:
        if isinstance(args, str):
            return json.loads(args)
        elif isinstance(args, dict):
            return args

    @model_validator(mode="after")
    def init_workspace_and_timestamp(self) -> Self:
        if not self.timestamp:
            self.timestamp = datetime.now().isoformat()
        self.workspace = self.workspace or os.environ.get("WORKSPACE", "")
        return self

    @field_serializer(*["completion_args", "extras"], when_used='json')
    def json_dump_completion_args(self, completion_args: Dict[str, Any]) -> str:
        return json.dumps(completion_args, indent=4)

    @property
    def messages(self) -> List[Dict[str, str]]:
        return self.completion_args.get("messages", [])

    @computed_field
    def model(self) -> str:
        return self.completion_args.get("model", "")

    @computed_field
    def temperature(self) -> float:
        return self.completion_args.get("temperature", 0.0)

    @computed_field
    def max_tokens(self) -> int:
        return self.completion_args.get("max_tokens", 0) or self.completion_args.get("max_completion_tokens", 0)

    @computed_field
    def system_prompt(self) -> Optional[str]:
        for msg in self.messages:
            if msg.get("role") == "system":
                return msg.get("content", "")
        # anthropic system messages
        if self.completion_args.get("system"):
            return self.completion_args.get("system")
                
        return ""

    @computed_field
    def assistant_message(self) -> Optional[str]:
        for msg in self.messages[::-1]:
            if msg.get("role") == "assistant":
                content = msg.get("content", "")
                if isinstance(content, list):
                    content = "\n".join([str(entry) for entry in content])
                return content
        return ""

    @computed_field
    def user_prompt(self) -> Optional[str]:
        for msg in self.messages[::-1]:
            if msg.get("role") == "user":
                content = msg.get("content", "")
                if isinstance(content, list):
                    content = "\n".join([str(entry) for entry in content])
                return content
        return ""

    @computed_field
    def history_messages(self) -> Optional[str]:
        return json.dumps([
            msg for msg in self.messages
            if msg.get("content") not in [
                self.system_prompt,
                self.assistant_message,
                self.user_prompt
            ]
        ], indent=4)

    @computed_field
    def total_tokens(self) -> int:
        return self.input_tokens + self.output_tokens

    @computed_field
    def success(self) -> bool:
        return bool(self.response)

    @model_serializer
    def serialize_model(self) -> Dict[str, Any]:
        """Ensure a cohesive field order during serialization."""
        return {
            "session_id": self.session_id,
            "workspace": self.workspace,
            "agent_id": self.agent_id,
            "action_id": self.action_id,
            "timestamp": self.timestamp,
            "operation_id": self.operation_id,
            "operation_type": self.operation_type,
            "provider": self.provider,
            "model": self.model,
            "system_prompt": self.system_prompt,
            "user_prompt": self.user_prompt,
            "response": self.response,
            "success": self.success,
            "assistant_message": self.assistant_message,
            "history_messages": self.history_messages,
            "temperature": self.temperature,
            "max_tokens": self.max_tokens,
            "input_tokens": self.input_tokens,
            "output_tokens": self.output_tokens,
            "cached_tokens": self.cached_tokens,
            "total_tokens": self.total_tokens,
            "cost": self.cost,
            "latency_ms": self.latency_ms,
            "error_message": self.error_message,
            "completion_args": json.dumps(self.completion_args, indent=4),
            "extras": json.dumps(self.extras)
        }
    

class LlmOperationCollector(RootModel):
    root: List[LlmOperationRecord] = []
    _storage_path: Optional[Union[str, Path]] = None
    _table_initialized: Optional[bool] = False
    _last_inserted_record: Optional[str] = None
    _engine: Optional[Any] = None
    _async_engine: Optional[Any] = None
    _session_factory: Optional[Any] = None
    _async_session_factory: Optional[Any] = None
    _is_sessions_initialized :set = set()

    @model_validator(mode="after")
    def init_dbsession(self) -> Self:
        try:
            from sqlalchemy import create_engine
            from sqlalchemy.orm import sessionmaker
            from sqlalchemy.ext.asyncio import create_async_engine, async_sessionmaker            
            from aicore.observability.models import Base
            from dotenv import load_dotenv
            load_dotenv()
            
            conn_str = os.environ.get("CONNECTION_STRING")
            async_conn_str = os.environ.get("ASYNC_CONNECTION_STRING")
            
            try:
                if conn_str:
                    self._engine = create_engine(conn_str)
                    self._session_factory = sessionmaker(bind=self._engine)
                    Base.metadata.create_all(self._engine)
                    self._table_initialized = True 
                
                # Async Engine
                if async_conn_str:
                    self._async_engine = create_async_engine(async_conn_str)
                    self._async_session_factory = async_sessionmaker(
                        bind=self._async_engine, 
                        expire_on_commit=False
                    )
                
            except Exception as e:
                _logger.logger.warning(f"Database connection failed: {str(e)}")

        except ModuleNotFoundError:
           _logger.logger.warning("pip install core-for-ai[sql] for sql integration and setup ASYNC_CONNECTION_STRING env var")
        
        return self
    
    async def create_tables(self):        
        from aicore.observability.models import Base
        if not self._async_engine:
            return
        
        try:            
            from aicore.observability.models import Base
            
        except ModuleNotFoundError:
             _logger.logger.warning("pip install core-for-ai[sql] for sql integration and setup ASYNC_CONNECTION_STRING env var")
        try:
            async with self._async_engine.begin() as conn:
                await conn.run_sync(Base.metadata.create_all)
                self._table_initialized = True
        finally:
            await self._async_engine.dispose()

    @property
    def storage_path(self) -> Optional[Union[str, Path]]:
        return self._storage_path

    @storage_path.setter
    def storage_path(self, value: Union[str, Path]):
        self._storage_path = value

    def _store_to_file(self, new_record: LlmOperationRecord) -> None:
        """Store a new record by appending it to the JSON file.
        Always maintains a valid JSON array format with proper closing bracket.
        """
        # Create directory if it doesn't exist
        if not os.path.exists(os.path.dirname(self.storage_path)):
            os.makedirs(os.path.dirname(self.storage_path), exist_ok=True)
        
        # Case 1: File doesn't exist or is empty
        if not os.path.exists(self.storage_path) or os.path.getsize(self.storage_path) == 0:
            # Create a new file with just this record
            with open(self.storage_path, 'w', encoding=DEFAULT_ENCODING) as f:
                f.write('[\n')
                f.write(json.dumps(new_record.model_dump(), indent=4))
                f.write('\n]')
            return
        
        # Case 2: File exists with content - need to append
        with open(self.storage_path, 'r+', encoding=DEFAULT_ENCODING) as f:
            # Seek to the position right before the closing bracket
            f.seek(0, os.SEEK_END)  # Go to end of file
            pos = f.tell()          # Get current position
            
            # Go backwards until we find the closing bracket
            while pos > 0:
                pos -= 1
                f.seek(pos)
                char = f.read(1)
                if char == ']':
                    # Found the closing bracket
                    break
            
            if pos <= 0:  # Sanity check - this shouldn't happen with valid JSON
                # File might be corrupted, handle accordingly
                # For simplicity, we'll overwrite with a new file
                with open(self.storage_path, 'w', encoding=DEFAULT_ENCODING) as f_new:
                    f_new.write('[\n')
                    f_new.write(json.dumps(new_record.model_dump(), indent=4))
                    f_new.write('\n]')
                return
            
            # Truncate the file at this position to remove the closing bracket
            f.seek(pos)
            f.truncate()
            
            # Now append the new record and closing bracket
            f.write(',\n')
            f.write(json.dumps(new_record.model_dump(), indent=4))
            f.write('\n]')

    def read_all_records(self) -> "LlmOperationCollector":
        """Read all records from the file.
        The file is always maintained in valid JSON format.
        """
        if not os.path.exists(self.storage_path) or os.path.getsize(self.storage_path) == 0:
            return LlmOperationCollector.model_construct(root=[])
        
        # File is always in valid JSON format, so we can read directly
        with open(self.storage_path, 'r', encoding=DEFAULT_ENCODING) as f:
            try:
                data = json.loads(f.read())
                records = LlmOperationCollector.model_construct(
                    root=[LlmOperationRecord(**kwargs) for kwargs in data]
                )
                return records
            except json.JSONDecodeError:
                # Handle potential corrupted file
                return LlmOperationCollector.model_construct(root=[])

    @staticmethod
    def _clean_completion_args(args: Dict[str, Any]) -> Dict[str, Any]:
        """Clean request arguments to remove sensitive information."""
        cleaned = args.copy()
        # Remove potentially sensitive information like API keys
        cleaned.pop("api_key", None)
        return cleaned

    @classmethod
    def fom_observable_storage_path(cls, storage_path: Optional[str] = None) -> "LlmOperationCollector":
        obj = cls()
        env_path = os.environ.get("OBSERVABILITY_DATA_DEFAULT_FILE")
        if storage_path:
            obj.storage_path = storage_path
        elif env_path:
            obj.storage_path = env_path
        else:
            obj.storage_path = Path(DEFAULT_OBSERVABILITY_DIR) / DEFAULT_OBSERVABILITY_FILE
        return obj

    @classmethod
    def polars_from_file(cls, storage_path: Optional[str] = None) -> "pl.DataFrame":  # noqa: F821
        obj = cls.fom_observable_storage_path(storage_path)
        if os.path.exists(obj.storage_path):
            with open(obj.storage_path, 'r', encoding=DEFAULT_ENCODING) as f:
                obj = cls(root=json.loads(f.read()))
        try:
            import polars as pl
            dicts = obj.model_dump()
            return pl.from_dicts(dicts) if dicts else pl.DataFrame()
        except ModuleNotFoundError:
            _logger.logger.warning("pip install -r requirements-dashboard.txt")
            return None
    
    def _handle_record(
        self,
        completion_args: Dict[str, Any],
        operation_type: Literal["completion", "acompletion"],
        provider: str,
        response: Optional[Union[str, Dict[str, str]]] = None,
        session_id: Optional[str] = None,
        workspace: Optional[str] = None,
        agent_id: Optional[str] = None,
        action_id: Optional[str] = None,
        input_tokens: Optional[int] = 0,
        output_tokens: Optional[int] = 0,
        cached_tokens: Optional[int] = 0,
        cost: Optional[float] = 0,
        latency_ms: Optional[float] = None,
        error_message: Optional[str] = None,
        extras: Optional[Dict[str, Any]] = None
    ) -> LlmOperationRecord:
        # Clean request args
        cleaned_args = self._clean_completion_args(completion_args)

        if not isinstance(response, (str, dict, list)) and response is not None:
            return None
        
        # Build a record
        record = LlmOperationRecord(
            session_id=session_id,
            agent_id=agent_id,
            action_id=action_id,
            workspace=workspace,
            provider=provider,
            operation_type=operation_type,
            input_tokens=input_tokens,
            output_tokens=output_tokens if not cached_tokens else output_tokens + cached_tokens,
            cached_tokens=cached_tokens,
            cost=cost,
            latency_ms=latency_ms or 0,
            error_message=error_message,
            completion_args=cleaned_args,
            response=response,
            extras=extras or {}
        )
        if self.storage_path:
            self._store_to_file(record)
        
        self.root.append(record)

        return record

    def record_completion(
        self,
        completion_args: Dict[str, Any],
        operation_type: Literal["completion", "acompletion"],
        provider: str,
        response: Optional[Union[str, Dict[str, str]]] = None,
        session_id: Optional[str] = None,
        workspace: Optional[str] = None,
        agent_id: Optional[str] = None,
        action_id: Optional[str] = None,
        input_tokens: Optional[int] = 0,
        output_tokens: Optional[int] = 0,
        cached_tokens: Optional[int] = 0,
        cost: Optional[float] = 0,
        latency_ms: Optional[float] = None,
        error_message: Optional[str] = None,
        extras: Optional[str] = None
    ) -> LlmOperationRecord:
        # Create record
        record = self._handle_record(
            completion_args, operation_type, provider, response, 
            session_id, workspace, agent_id, action_id, 
            input_tokens, output_tokens, cached_tokens, cost, latency_ms, error_message, extras
        )
        
        if self._engine and self._session_factory and record:
            try:
                self._insert_record_to_db(record)
            except Exception as e:
                _logger.logger.error(f"Error inserting record to DB: {str(e)}")
        
        return record
    
    async def arecord_completion(
        self,
        completion_args: Dict[str, Any],
        operation_type: Literal["completion", "acompletion"],
        provider: str,
        response: Optional[Union[str, Dict[str, str]]] = None,
        session_id: Optional[str] = None,
        workspace: Optional[str] = None,
        agent_id: Optional[str] = None,
        action_id: Optional[str] = None,
        input_tokens: Optional[int] = 0,
        output_tokens: Optional[int] = 0,
        cached_tokens: Optional[int] = 0,
        cost: Optional[float] = 0,
        latency_ms: Optional[float] = None,
        error_message: Optional[str] = None,
        extras: Optional[str] = None
    ) -> LlmOperationRecord:
        # Create record
        record = self._handle_record(
            completion_args, operation_type, provider, response, 
            session_id, workspace, agent_id, action_id, 
            input_tokens, output_tokens, cached_tokens, cost, latency_ms, error_message, extras
        )
        
        if self._async_engine and self._async_session_factory and record:
            if not self._table_initialized:
                await self.create_tables()
            try:
                await self._a_insert_record_to_db(record)
            except Exception as e:
                _logger.logger.error(f"Error inserting record to DB: {str(e)}")
        
        return record
    
    def _insert_record_to_db(self, record: LlmOperationRecord) -> None:
        """Insert a single LLM operation record into the database using SQLAlchemy."""
        try:
            from aicore.observability.models import Session, Message, Metric
        except ModuleNotFoundError:
             _logger.logger.warning("pip install core-for-ai[sql] for sql integration and setup ASYNC_CONNECTION_STRING env var")
        
        if not self._session_factory:
            if self._async_session_factory:
                _logger.logger.warning("You have configured an async connection to a db but are trying to establish a sync one. Pass CONNECTION_STRING env var.")
            return
            
        serialized = record.serialize_model()
        
        # Use context manager for session handling
        with self._session_factory() as session:
            try:
                # Check if session exists, create if it doesn't
                db_session_query = session.query(Session).filter_by(session_id=serialized['session_id'])
                db_session = list(db_session_query.all())  # Force fetch all results
                
                if not db_session:
                    db_session = Session(
                        session_id=serialized['session_id'],
                        workspace=serialized['workspace'],
                        agent_id=serialized['agent_id']
                    )
                    session.add(db_session)
                    session.flush()  # Flush changes to DB but don't commit yet
                else:
                    db_session = db_session[0]  # Get first result
                
                # Create message record
                message = Message(
                    operation_id=serialized['operation_id'],
                    session_id=serialized['session_id'],
                    action_id=serialized['action_id'],
                    timestamp=serialized['timestamp'],
                    system_prompt=serialized['system_prompt'],
                    user_prompt=serialized['user_prompt'],
                    response=serialized['response'],
                    assistant_message=serialized['assistant_message'],
                    history_messages=serialized['history_messages'],
                    completion_args=serialized['completion_args'],
                    error_message=serialized['error_message']
                )
                session.add(message)
                
                # Create metrics record
                metric = Metric(
                    operation_id=serialized['operation_id'],
                    operation_type=serialized['operation_type'],
                    provider=serialized['provider'],
                    model=serialized['model'],
                    success=serialized['success'],
                    temperature=serialized['temperature'],
                    max_tokens=serialized['max_tokens'],
                    input_tokens=serialized['input_tokens'],
                    output_tokens=serialized['output_tokens'],
                    cached_tokens=serialized['cached_tokens'],
                    total_tokens=serialized['total_tokens'],
                    cost=serialized['cost'],
                    latency_ms=serialized['latency_ms'],
                    extras=serialized['extras']
                )
                session.add(metric)
                
                # Commit all changes
                session.commit()
                self._last_inserted_record = serialized['operation_id']
            except Exception as e:
                session.rollback()
                raise e

    async def _a_insert_record_to_db(self, record: LlmOperationRecord) -> None:
        """Insert a single LLM operation record into the database asynchronously."""
        if not self._async_session_factory:
            if self._session_factory:
                _logger.logger.warning("You have configured a sync connection to a db but are trying to establish an async one. Pass ASYNC_CONNECTION_STRING env var.")
            return

        serialized = record.serialize_model()
        
        try:
            from sqlalchemy.future import select
            from aicore.observability.models import Session, Message, Metric
            
            # First with block: Handle session creation
            async with self._async_session_factory() as session:
                try:
                    # Check if session exists, create if it doesn't
                    if serialized['session_id'] not in self._is_sessions_initialized:
                        # Check if any row exists
                        db_session = await session.scalar(
                            select(1).where(Session.session_id == serialized['session_id'])
                        )
                        result = db_session is not None

                        if not result:
                            db_session = Session(
                                session_id=serialized['session_id'],
                                workspace=serialized['workspace'],
                                agent_id=serialized['agent_id']
                            )
                            session.add(db_session)
                            await session.commit()
                            self._is_sessions_initialized.add(serialized['session_id'])
                except Exception as e:
                    await session.rollback()
                    raise e

            # Second with block: Handle message insertion
            async with self._async_session_factory() as session:
                try:
                    message = Message(
                        operation_id=serialized['operation_id'],
                        session_id=serialized['session_id'],
                        action_id=serialized['action_id'],
                        timestamp=serialized['timestamp'],
                        system_prompt=serialized['system_prompt'],
                        user_prompt=serialized['user_prompt'],
                        response=serialized['response'],
                        assistant_message=serialized['assistant_message'],
                        history_messages=serialized['history_messages'],
                        completion_args=serialized['completion_args'],
                        error_message=serialized['error_message']
                    )
                    session.add(message)
                    await session.commit()
                except Exception as e:
                    await session.rollback()
                    raise e

            # Third with block: Handle metrics insertion
            async with self._async_session_factory() as session:
                try:
                    metric = Metric(
                        operation_id=serialized['operation_id'],
                        operation_type=serialized['operation_type'],
                        provider=serialized['provider'],
                        model=serialized['model'],
                        success=serialized['success'],
                        temperature=serialized['temperature'],
                        max_tokens=serialized['max_tokens'],
                        input_tokens=serialized['input_tokens'],
                        output_tokens=serialized['output_tokens'],
                        cached_tokens=serialized['cached_tokens'],
                        total_tokens=serialized['total_tokens'],
                        cost=serialized['cost'],
                        latency_ms=serialized['latency_ms'],
                        extras=serialized['extras']
                    )
                    session.add(metric)
                    await session.commit()
                except Exception as e:
                    await session.rollback()
                    raise e
            
            # Set the last inserted record after all operations succeed
            self._last_inserted_record = serialized['operation_id']
            
        finally:
            await self._async_engine.dispose()

    @classmethod
    def polars_from_db(cls,
        agent_id: Optional[str] = None,
        action_id: Optional[str] = None,
        session_id: Optional[str] = None,
        workspace: Optional[str] = None,
        start_date: Optional[str] = None,
        end_date: Optional[str] = None
    ) -> "pl.DataFrame":  # noqa: F821
        """
        Query the database and return results as a Polars DataFrame.
        Works with any database supported by SQLAlchemy.
        
        Defaults:
            - start_date: Midnight of the previous day
            - end_date: Now (current time)
        """
        # Set default start_date to midnight of the previous day
        if start_date is None:
            yesterday = datetime.now() - timedelta(days=1)
            start_date = yesterday.replace(hour=0, minute=0, second=0, microsecond=0).isoformat()

        # Set default end_date to now (optional)
        if end_date is None:
            end_date = datetime.now().isoformat()

        instance = cls()
        
        if instance._session_factory and instance._engine:
            return instance._polars_from_db(
                agent_id, action_id, session_id, workspace, start_date, end_date
            )
        elif instance._async_session_factory and instance._async_engine:
            coro = instance._apolars_from_db(agent_id, action_id, session_id, workspace, start_date, end_date)

            try:
                loop = asyncio.get_running_loop()
            except RuntimeError:
                # No running loop: safe to use asyncio.run
                return asyncio.run(coro)
            else:
                # Already inside a running loop — use `ensure_future` or `create_task`
                future = asyncio.ensure_future(coro)
                return asyncio.get_event_loop().run_until_complete(future)
        else:
            try:
                import polars as pl
                return pl.DataFrame()
            except ModuleNotFoundError:
                _logger.logger.warning("pip install core-for-ai[all] for Polars and sql integration")
                return None

    def _polars_from_db(self,
                    agent_id: Optional[str] = None,
                    action_id: Optional[str] = None,
                    session_id: Optional[str] = None,
                    workspace: Optional[str] = None,
                    start_date: Optional[str] = None,
                    end_date: Optional[str] = None) -> "pl.DataFrame":  # noqa: F821
        """
        Query the database (using SQLAlchemy) and return results as a Polars DataFrame.
        Works with any database supported by SQLAlchemy.
        """
        try:
            import polars as pl
            from sqlalchemy import desc
            from aicore.observability.models import Session, Message, Metric
        except ModuleNotFoundError:
            _logger.logger.warning("pip install core-for-ai[all] for Polars and sql integration")
            return None
        
        with self._session_factory() as session:
            try:
                # Build query with filters
                query = session.query(
                    Session.session_id, Session.workspace, Session.agent_id,
                    Message.action_id, Message.operation_id, Message.timestamp, 
                    Message.system_prompt, Message.user_prompt, Message.response,
                    Message.assistant_message, Message.history_messages, 
                    Message.completion_args, Message.error_message,
                    Metric.operation_type, Metric.provider, Metric.model, 
                    Metric.success, Metric.temperature, Metric.max_tokens, 
                    Metric.input_tokens, Metric.output_tokens, Metric.cached_tokens, Metric.total_tokens,
                    Metric.cost, Metric.latency_ms, Metric.extras
                ).join(
                    Message, Session.session_id == Message.session_id
                ).join(
                    Metric, Message.operation_id == Metric.operation_id
                ).filter(
                    Message.timestamp >= start_date,
                    Message.timestamp <= end_date
                )
                
                # Apply filters
                if agent_id:
                    query = query.filter(Session.agent_id == agent_id)
                if action_id:
                    query = query.filter(Message.action_id == action_id)
                if session_id:
                    query = query.filter(Session.session_id == session_id)
                if workspace:
                    query = query.filter(Session.workspace == workspace)
                if start_date:
                    query = query.filter(Message.timestamp >= start_date)
                if end_date:
                    query = query.filter(Message.timestamp <= end_date)
                    
                # Order by operation_id descending
                query = query.order_by(desc(Message.operation_id))
                
                # Force immediate consumption of all results to prevent "Connection is busy" errors
                results = list(query.all())
                
                if not results:
                    return pl.DataFrame()
                
                # Convert to dictionary
                records = []
                for row in results:
                    record = {}
                    for idx, column in enumerate(query.column_descriptions):
                        record[column['name']] = row[idx]
                    records.append(record)
                    
                # Ensure session is clean before returning
                session.commit()
                
                # Convert to Polars DataFrame
                return pl.from_dicts(records)
                
            except Exception as e:
                _logger.logger.warning(f"collector.py:640 Error executing database query: {str(e)}")
                session.rollback()  # Explicitly rollback on error
                return pl.DataFrame()
   
    async def _apolars_from_db(self,
        agent_id: Optional[str] = None,
        action_id: Optional[str] = None,
        session_id: Optional[str] = None,
        workspace: Optional[str] = None,
        start_date: Optional[str] = None,
        end_date: Optional[str] = None
    ) -> "pl.DataFrame":  # noqa: F821
        """
        Query the database asynchronously (using SQLAlchemy) and return results as a Polars DataFrame.
        """
        try:
            import polars as pl
            from sqlalchemy import desc, select
            from aicore.observability.models import Session, Message, Metric
        except ModuleNotFoundError:
            _logger.logger.warning("pip install core-for-ai[all] for Polars and sql integration")
            return None
        
        async with self._async_session_factory() as session:
            try:
                session = self._async_session_factory()
                query = (
                    select(
                        Session.session_id, Session.workspace, Session.agent_id,
                        Message.action_id, Message.operation_id, Message.timestamp, 
                        Message.system_prompt, Message.user_prompt, Message.response,
                        Message.assistant_message, Message.history_messages, 
                        Message.completion_args, Message.error_message,
                        Metric.operation_type, Metric.provider, Metric.model, 
                        Metric.success, Metric.temperature, Metric.max_tokens, 
                        Metric.input_tokens, Metric.output_tokens, Metric.cached_tokens, Metric.total_tokens,
                        Metric.cost, Metric.latency_ms, Metric.extras
                    )
                    .join(Message, Session.session_id == Message.session_id)
                    .join(Metric, Message.operation_id == Metric.operation_id)
                    .filter(
                        Message.timestamp >= start_date,
                        Message.timestamp <= end_date
                    )
                )

                # Apply filters
                if agent_id:
                    query = query.where(Session.agent_id == agent_id)
                if action_id:
                    query = query.where(Message.action_id == action_id)
                if session_id:
                    query = query.where(Session.session_id == session_id)
                if workspace:
                    query = query.where(Session.workspace == workspace)
                
                query = query.order_by(desc(Message.operation_id))

                # Execute query and immediately consume all results
                result = await session.execute(query)
                rows = result.fetchall()  # eager fetch
                
                # Explicitly commit to ensure connection is cleared
                await session.commit()

                if not rows:
                    return pl.DataFrame()
                
                # Convert to dictionary
                records = [dict(row._asdict()) for row in rows]
                return pl.from_dicts(records)
            except Exception as e:
                _logger.logger.error(f"Error executing database query: {str(e)}")
                await session.rollback()  # Explicitly rollback on error
                return pl.DataFrame()            
            
            finally:
                await self._async_engine.dispose()

if __name__ == "__main__":
    LlmOperationCollector()
    df = LlmOperationCollector.polars_from_db()
    print(df.columns)
    print(df)


from aicore.models_metadata import PricingConfig

from pydantic import BaseModel, RootModel, Field, computed_field, model_validator
from typing import Optional, List, Union
from datetime import datetime, timezone
from collections import defaultdict
from ulid import ulid

class CompletionUsage(BaseModel):
    """Tracks token usage and cost for a single LLM completion."""
    completion_id: Optional[str] = Field(default_factory=ulid)
    prompt_tokens: int
    response_tokens: int
    cached_tokens: int = 0
    cache_write_tokens: int = 0
    cost: Optional[float] = 0

    @property
    def input_tokens(self) -> int:
        """Returns the number of input/prompt tokens."""
        return self.prompt_tokens
    
    @property
    def output_tokens(self) -> int:
        """Returns the number of output/response tokens."""
        return self.response_tokens

    @model_validator(mode="after")
    def validate_token_counts(self) -> "CompletionUsage":
        """Validate that token counts are non-negative."""
        if self.prompt_tokens < 0 or self.response_tokens < 0:
            raise ValueError("Token counts cannot be negative")
        return self
    
    @computed_field
    def total_tokens(self) -> int:
        """Returns the total number of tokens used (input + output)."""
        return self.input_tokens + self.output_tokens
    
    def __str__(self) -> str:
        """Returns a human-readable string representation of the usage."""
        cost_prefix = f"Cost: ${self.cost} | " if self.cost else ""
        return f"{cost_prefix}Tokens: {self.total_tokens} | Prompt: {self.prompt_tokens} | Response: {self.response_tokens}"
    
    @classmethod
    def from_pricing_info(
        cls,
        completion_id: str,
        prompt_tokens: int,
        response_tokens: int,
        cached_tokens: int = 0,
        cache_write_tokens: int = 0,
        cost: Optional[float] = 0,
        pricing: Optional[PricingConfig] = None
    ) -> "CompletionUsage":
        """Creates a CompletionUsage instance with calculated cost based on pricing config."""
        total_input_tokens = prompt_tokens + cached_tokens + cache_write_tokens
        if pricing is not None:
            # Apply happy hour pricing if active
            if pricing.happy_hour is not None and pricing.happy_hour.start <= datetime.now(timezone.utc) <= pricing.happy_hour.finish:
                pricing = pricing.happy_hour.pricing
            
            input_cost = pricing.input * prompt_tokens
            output_cost = pricing.output * response_tokens
            cached_cost = pricing.cached * cached_tokens
            cache_write_cost = cache_write_tokens * pricing.cache_write

            # Apply dynamic pricing only if threshold is exceeded
            if pricing.dynamic is not None:
                total_tokens = total_input_tokens + response_tokens
                if total_tokens > pricing.dynamic.threshold:
                    # Calculate how many tokens are over the threshold
                    tokens_over_threshold = total_tokens - pricing.dynamic.threshold
                    
                    # First, determine how many input tokens are over the threshold
                    input_tokens_over_threshold = max(0, prompt_tokens - pricing.dynamic.threshold)
                    
                    # Then calculate remaining tokens over threshold (must be output tokens)
                    output_tokens_over_threshold = tokens_over_threshold - input_tokens_over_threshold
                    
                    # Recalculate costs with separate pricing for tokens above and below threshold
                    input_cost = (
                        pricing.input * (prompt_tokens - input_tokens_over_threshold) +  # Base price for tokens <= threshold
                        pricing.dynamic.pricing.input * input_tokens_over_threshold      # Dynamic price for excess tokens
                    )
                    
                    output_cost = (
                        pricing.output * (response_tokens - output_tokens_over_threshold) +  # Base price for tokens <= threshold
                        pricing.dynamic.pricing.output * output_tokens_over_threshold        # Dynamic price for excess tokens
                    )

            # Final cost calculation (always includes cached/cache_write costs)
            cost = (input_cost + output_cost + cached_cost + cache_write_cost) * 1e-6

        return cls(
            completion_id=completion_id,
            prompt_tokens=total_input_tokens,
            response_tokens=response_tokens,
            cached_tokens=cached_tokens,
            cache_write_tokens=cache_write_tokens,
            cost=cost,
        )
    
    def update_with_pricing(self, pricing: PricingConfig):
        """Updates the cost based on the given pricing config if not already set. Does not take into account cost of cache writing"""
        if not self.cost:
            if pricing.happy_hour is not None and pricing.happy_hour.start <= datetime.now(timezone.utc) <= pricing.happy_hour.finish:
                pricing = pricing.happy_hour.pricing
            
            if pricing.dynamic is not None and self.prompt_tokens + self.response_tokens > pricing.dynamic.threshold:
                pricing = pricing.dynamic.pricing
            
            self.cost = (
                pricing.input * self.prompt_tokens 
                + pricing.output * self.response_tokens
                + pricing.cached * self.cached_tokens
                + pricing.cache_write * self.cache_write_tokens
            ) * 1e-6

class UsageInfo(RootModel):
    """Aggregates token usage and cost across multiple completions."""
    root: List[CompletionUsage] = []
    _pricing: Optional[PricingConfig] = None
    _allow_negative_costs: bool = False

    @classmethod
    def from_pricing_config(cls, pricing: PricingConfig) -> "UsageInfo":
        """Creates a UsageInfo instance with the given pricing config."""
        instance = cls()
        instance.pricing = pricing
        return instance

    @property
    def allow_negative_costs(self) -> bool:
        """Returns whether negative costs are allowed (for testing purposes)."""
        return self._allow_negative_costs

    def record_completion(
        self,
        prompt_tokens: int,
        response_tokens: int,
        cached_tokens: int = 0,
        cache_write_tokens: int = 0,
        completion_id: Optional[str] = None
    ):
        """Records a new completion with the given token counts."""
        if completion_id is None and self.root:
            completion_id = self.latest_completion.completion_id
            
        self.root.append(CompletionUsage.from_pricing_info(
            completion_id=completion_id,
            prompt_tokens=prompt_tokens,
            cached_tokens=cached_tokens,
            cache_write_tokens=cache_write_tokens,
            response_tokens=response_tokens,
            pricing=self.pricing
        ))

    @computed_field
    def pricing(self) -> Optional[PricingConfig]:
        """Returns the current pricing configuration."""
        return self._pricing
    
    @pricing.setter
    def pricing(self, pricing_info: PricingConfig):
        """Sets the pricing configuration."""
        self._pricing = pricing_info

    def set_pricing(self, input_1m: float, output_1m: float):
        """Sets basic pricing rates (per 1M tokens)."""
        self._pricing = PricingConfig(
            input=input_1m,
            output=output_1m
        )

    @computed_field
    def latest_completion(self) -> Union[None, CompletionUsage]:
        """Returns the most recent completion."""
        return self.completions[-1] if self.root else None
    
    def _is_aggregated(self) -> bool:
        """Checks if completions are already aggregated by completion_id."""
        completion_ids = [item.completion_id for item in self.root]
        return len(completion_ids) == len(set(completion_ids))

    @computed_field
    def completions(self) -> List[CompletionUsage]:
        """Returns aggregated completions by completion_id."""
        if self._is_aggregated():
            return self.root
        
        # Dictionary to store the final aggregated values for each completion_id
        aggregated_completions = {}        
        # Group all items by completion_id
        completion_groups = defaultdict(list)

        for item in self.root:
            completion_groups[item.completion_id].append(item)
        
        # Process each completion_id group
        for comp_id, items in completion_groups.items():
            # Group items by prompt_tokens to handle duplicates properly
            prompt_token_groups = defaultdict(list)
            for item in items:
                prompt_token_groups[item.prompt_tokens].append(item)
            # Take the sum of prompt_tokens (should be the same for all items in a group)
            prompt_tokens = next(iter(prompt_token_groups.keys())) if prompt_token_groups else 0
            # Process response tokens based on prompt token uniqueness
            total_response_tokens = 0
            if len(prompt_token_groups) > 1:
                # If we have different prompt_tokens values, sum all response tokens
                total_response_tokens = sum(item.response_tokens for item in items)
            else:
                # If all prompt_tokens are equal, for each prompt_token value, 
                # take only the maximum response_tokens
                for pt_value, pt_items in prompt_token_groups.items():
                    max_response_tokens = max(item.response_tokens for item in pt_items)
                    total_response_tokens += max_response_tokens
            
            # For cached_tokens and cache_write_tokens, take the maximum value
            total_cached_tokens = max((item.cached_tokens for item in items), default=0)
            total_cache_write_tokens = max((item.cache_write_tokens for item in items), default=0)
            
            # Create a single aggregated CompletionUsage object
            aggregated_completions[comp_id] = CompletionUsage.from_pricing_info(
                completion_id=comp_id,
                prompt_tokens=prompt_tokens,
                response_tokens=total_response_tokens,
                cached_tokens=total_cached_tokens,
                cache_write_tokens=total_cache_write_tokens,
                pricing=self.pricing
            )
        
        # Replace the root with the aggregated completions
        self.root = list(aggregated_completions.values())
        
        return self.root
    
    @computed_field
    def total_tokens(self) -> int:
        """Returns the total number of tokens across all completions."""
        return sum(completion.total_tokens for completion in self.completions)
    
    @computed_field
    def total_cost(self) -> float:
        """Returns the total cost across all completions."""
        return sum(completion.cost for completion in self.completions)
    
    def __str__(self) -> str:
        """Returns a human-readable string representation of the total usage."""
        cost_prefix = f"Cost: ${self.total_cost} | " if self.total_cost else ""
        return f"Total | {cost_prefix} Tokens: {self.total_tokens}"



from typing import Dict, List, Optional, Any, AsyncContextManager, Union
from pydantic import BaseModel, ConfigDict, Field
from pathlib import Path
import asyncio
import time
import json
import os

from fastmcp import Client as FastMCPClient
from fastmcp import FastMCP as FastMCPServer
from fastmcp.client.transports import (
    FastMCPTransport,
    WSTransport, 
    SSETransport,
    StdioTransport,
    StdioServerParameters
)

from aicore.llm.mcp.models import MCPParameters, MCPServerConfig, SSSEParameters, ToolSchema, WSParameters
from aicore.llm.mcp.utils import raise_fast_mcp_error
from aicore.logger import _logger

class ServerConnection(AsyncContextManager):
    """
    A wrapper around FastMCPClient that implements the AsyncContextManager protocol.
    This allows users to access server connections using async with statements.
    """
    def __init__(self, transport: FastMCPTransport):
        self.transport = transport
        self.client = None
        
    async def __aenter__(self):
        self.client = FastMCPClient(self.transport)
        await self.client.__aenter__()
        return self.client
        
    async def __aexit__(self, exc_type, exc_val, exc_tb):
        if self.client:
            await self.client.__aexit__(exc_type, exc_val, exc_tb)
            
class ServerManager:
    """
    Manages server connections and provides convenient access to server clients.
    """
    def __init__(self, parent_client: 'MCPClient'):
        self._parent = parent_client
        self._tools_cache = {}  # Cache for mapping tool names to server names
        self._servers_cache :Dict[str, ToolSchema]= {}
        
    def get(self, server_name: str) -> ServerConnection:
        """Get a connection to a specific server by name."""
        if server_name not in self._parent.transports:
            raise KeyError(f"Server '{server_name}' not found in configured servers")
        
        return ServerConnection(self._parent.transports[server_name])
    
    @raise_fast_mcp_error(prefix="mcp-get-tools")
    async def get_tools(self) -> Dict[str, List[ToolSchema]]:
        """Get all tools from all connected servers as a dictionary mapping server names to tool lists."""
        result = {}
        
        for server_name, transport in self._parent.transports.items():
            async with FastMCPClient(transport) as client:
                try:
                    tools = await client.list_tools()
                    result[server_name] = [
                        ToolSchema.from_mcp_tool(tool)
                        for tool in tools
                    ]
                except Exception as e:
                    result[server_name] = [f"Error: {str(e)}"]
                    
        return result
    
    async def get_servers(self) -> Dict[str, str]:
        all_tools =  await self.get_tools()
        for client, servers in all_tools.items():
            for server in servers:
                self._servers_cache[server.name] = client

    @property
    async def tools(self) -> List[ToolSchema]:
        """Get a flat list of all tools from all connected servers."""
        tools_by_server = await self.get_tools()
        all_tools = []
        
        for server_name, tools in tools_by_server.items():
            # Filter out error messages which start with "Error:"
            valid_tools = [tool for tool in tools if not isinstance(tool, str) or not tool.startswith("Error:")]
            all_tools.extend(valid_tools)
            
        return all_tools
    
    @raise_fast_mcp_error(prefix="mcp-call-tool")
    async def call_tool(self, tool_name: str, arguments: Any = None, silent :Optional[bool]=False) -> Any:
        """
        Call a tool by name without specifying which server it belongs to.
        The system will automatically determine which server provides the tool.
        
        Args:
            tool_name: The name of the tool to call
            arguments: The arguments to pass to the tool
            
        Returns:
            The result of the tool call
            
        Raises:
            ValueError: If the tool is not found on any server
            ValueError: If the tool is found on multiple servers
        """
        # If the tool cache is empty, populate it
        if not self._tools_cache:
            await self.get_servers()
            
        # If the tool is still not in the cache, it's not available
        if tool_name not in self._servers_cache:
            raise ValueError(f"Tool '{tool_name}' not found on any connected server")
        
        server_name = self._servers_cache[tool_name]
        
        # Call the tool on the appropriate server
        _logger.logger.info(f"MCP | Starting call to tool '{tool_name}' on server '{server_name}' with arguments: {arguments}") if not silent else ...
        st = time.perf_counter()
        async with self.get(server_name) as client:
            result = await client.call_tool(tool_name, arguments)
        duration = time.perf_counter() - st
        _logger.logger.info(f"MCP | Finished call to tool '{tool_name}' on server '{server_name}' in {duration:.2f}s") if not silent else ...
        return result

class MCPClient(BaseModel):
    """
    MCP Client that can connect to one or more MCP servers defined in a config file.
    One client can have multiple connections to different servers.
    All interactions should be done within a context manager (with block).
    """
    server_configs: Dict[str, MCPServerConfig] = Field(default_factory=dict)
    transports: Dict[str, FastMCPTransport] = Field(default_factory=dict, exclude=True)
    _is_connected: bool=False
    _needs_update: bool=False
    
    model_config = ConfigDict(
        arbitrary_types_allowed=True
    )
    
    def __init__(self, **data: Any):
        super().__init__(**data)
        # Initialize the server manager
        self._servers = ServerManager(self)
    
    @property
    def servers(self) -> ServerManager:
        """Access the server manager."""
        return self._servers
    
    @classmethod
    def from_config(cls, config: Union[str, Path, Dict[str, str]]) -> "MCPClient":
        """Create an MCPClient instance from a config file."""
        if not isinstance(config, dict):
            if not os.path.exists(config):
                raise FileNotFoundError(f"MCP config file not found: {config}")
            
            with open(config, "r") as f:
                config = json.load(f)
        
        client = cls()
        
        if "mcpServers" in config and isinstance(config["mcpServers"], dict):
            for server_name, server_config in config["mcpServers"].items():
                transport_type = server_config.get("transport_type", "stdio")
                client.server_configs[server_name] = MCPServerConfig(
                    name=server_name,
                    parameters=getattr(MCPParameters, transport_type).value(**server_config),
                    transport_type=transport_type,
                    additional_params=server_config.get("additional_params", {})
                )
        
        return client
    
    def add_server(self, name: str, parameters :Union[FastMCPServer, StdioServerParameters, SSSEParameters, WSParameters], **kwargs) -> None:
        """Add a server configuration manually."""        
        self.server_configs[name] = MCPServerConfig(
            name=name,
            parameters=parameters,
            additional_params=kwargs
        )
        self.transports[name] = self._create_transport(self.server_configs[name])
        self._needs_update = True
    
    @raise_fast_mcp_error(prefix="mcp-connect")
    async def connect(self, server_name: Optional[str] = None) -> None:
        """
        Connect to one or all configured MCP servers.
        
        Args:
            server_name: If provided, connect only to this specific server.
                         If None, connect to all configured servers.
        """
        server_names = [server_name] if server_name else list(self.server_configs.keys())
        for name in server_names:
            if name not in self.server_configs:
                raise KeyError(f"Server '{name}' not found in configured servers")
                
            server_config = self.server_configs[name]
            transport = self._create_transport(server_config)
            self.transports[name] = transport
        self._is_connected = True
    
    def _create_transport(self, server_config: MCPServerConfig) -> FastMCPTransport:
        """Create an appropriate transport based on the server configuration."""
        params = server_config.parameters
        
        if isinstance(params, FastMCPServer):
            return FastMCPTransport(params)
        elif isinstance(params, StdioServerParameters):
            return StdioTransport(
                command=params.command,
                args=params.args,
                env=params.env,
                cwd=params.cwd
            )
        elif isinstance(params, WSParameters):
            return WSTransport(url=params.url)
        elif isinstance(params, SSSEParameters):
            return SSETransport(url=params.url, headers=params.headers)
        else:
            raise ValueError("Unsupported transport type")
            
    async def __aenter__(self) -> "MCPClient":
        """Connect to all servers when entering the context manager."""
        await self.connect()
        return self
        
    async def __aexit__(self, exc_type, exc_val, exc_tb):
        """Clean up resources when exiting the context manager."""
        self.transports.clear()


# Example usage:
async def example_usage():
    # Create from config file
    mcp = MCPClient.from_config_file("mcp_config.json")
    
    # Connect to all servers
    await mcp.connect()
    
    # Option 1: Use the ServerManager to get a specific server
    async with mcp.servers.get("brave-search") as brave_client:
        tools = await brave_client.list_tools()
        print(f"Tools from brave-search: {tools}")
        
        # Call a tool on a specific server
        result = await brave_client.call_tool("brave_web_search", {"query": "python programming"})
        print(f"Search result: {result}")
    
    # Option 2: Get tools by server (dictionary format)
    tools_by_server = await mcp.servers.get_tools()
    print(f"Tools organized by server: {tools_by_server}")
    
    # Option 3: Get flat list of all tools from all servers
    all_tools = await mcp.servers.tools
    print(f"All tools (flat list): {all_tools}")
    
    # Option 4: Call a tool without specifying the server
    try:
        tool_result = await mcp.servers.call_tool("brave_web_search", {"query": "artificial intelligence"})
        print(f"Tool result: {tool_result}")
    except ValueError as e:
        print(f"Error calling tool: {e}")
    
    # Use with context manager
    async with MCPClient.from_config_file("mcp_config.json") as mcp:
        # All servers are automatically connected
        
        # Get flat list of all tools
        all_tools = await mcp.servers.tools
        print(f"All available tools: {all_tools}")
        
        # Call a tool without specifying the server
        if "brave_web_search" in all_tools:
            result = await mcp.servers.call_tool("brave_web_search", {"query": "machine learning"})
            print(f"Search result: {result}")

if __name__ == "__main__":
    asyncio.run(example_usage())



from fastmcp.client.transports import StdioServerParameters
from fastmcp import FastMCP as FastMCPServer
from mcp.types import Tool

from typing import Any, Dict, List, Literal, Optional, Union
from pydantic import BaseModel, ConfigDict, Field, RootModel
from enum import Enum

class WSParameters(BaseModel):
    url: str

class SSSEParameters(BaseModel):
    url: str
    headers: dict[str, Any] | None = None
    timeout: float = 5
    sse_read_timeout: float = 60 * 5

class MCPServerConfig(BaseModel):
    """Configuration for a single MCP server connection."""
    name: str
    parameters :Union[FastMCPServer, StdioServerParameters, SSSEParameters, WSParameters]
    transport_type: Literal["stdio", "sse", "ws"] = "stdio"  # Can be "fastmcp", "stdio", "ws", "sse", or None for stdio
    additional_params: Dict[str, Any] = Field(default_factory=dict)

    model_config = ConfigDict(
        arbitrary_types_allowed=True
    )

class MCPParameters(Enum):
    ws :WSParameters=WSParameters
    sse :SSSEParameters=SSSEParameters
    stdio :StdioServerParameters=StdioServerParameters

class PropertiesSchema(BaseModel):

    model_config = ConfigDict(
        extra="allow",
    )

class InputSchema(BaseModel):
    type :str
    properties :PropertiesSchema
    required :List[str]

    model_config = ConfigDict(
        extra="allow",
    )

class ToolSchema(BaseModel):
    name :str
    description :str
    input_schema :InputSchema

    @classmethod
    def from_mcp_tool(cls, tool :Tool)->"ToolSchema":
        return cls(
            name=tool.name,
            description=tool.description,
            input_schema=InputSchema(**tool.inputSchema)
        )

class ToolCallSchema(BaseModel):
    type :str="function"
    id :str
    name :str
    arguments :Union[str, Dict]
    _raw :Optional[Any]=None
    

    """
    [
        Tool(
            name='brave_web_search',
            description='Performs a web search using the Brave Search API, ideal for general queries, news, articles, and online content.
                Use this for broad information gathering, recent events, or when you need diverse web sources.
                Supports pagination, content filtering, and freshness controls. Maximum 20 results per request, with offset for pagination.',
            inputSchema={
                'type': 'object',
                'properties': {
                    'query': {
                        'type': 'string',
                        'description': 'Search query (max 400 chars, 50 words)'
                    },
                    'count': {
                        'type': 'number',
                        'description': 'Number of results (1-20, default 10)', 'default': 10
                    },
                    'offset': {
                        'type': 'number',
                        'description': 'Pagination offset (max 9, default 0)',
                        'default': 0
                    }
                }, 
                'required': ['query']
            }
        )
    ]
    """

    """
    ### Anthropic
    {
        "name": "get_weather",
        "description": "Get the current weather in a given location",
        "input_schema": {
          "type": "object",
          "properties": {
            "location": {
              "type": "string",
              "description": "The city and state, e.g. San Francisco, CA"
            }
          },
          "required": [
            "location"
          ]
        }
      }
    ### Deepseek
    {
        "type": "function",
        "function": {
            "name": "get_weather",
            "description": "Get weather of an location, the user shoud supply a location first",
            "parameters": {
                "type": "object",
                "properties": {
                    "location": {
                        "type": "string",
                        "description": "The city and state, e.g. San Francisco, CA",
                    }
                },
                "required": ["location"]
            },
        }
    }
    #### OpenAI
    {
        "type": "function",
        "name": "get_weather",
        "description": "Get current temperature for a given location.",
        "parameters": {
            "type": "object",
            "properties": {
                "location": {
                    "type": "string",
                    "description": "City and country e.g. Bogotá, Colombia"
                }
            },
            "required": [
                "location"
            ],
            "additionalProperties": False
        }
    }
    """

class ToolCalls(RootModel):
    root: List[ToolCallSchema] = []


from aicore.embeddings.providers.openai import OpenAiEmbeddings

class GeminiEmbeddings(OpenAiEmbeddings):
    vector_dimensions :int=768
    base_url :str="https://generativelanguage.googleapis.com/v1beta/openai/"


from aicore.embeddings.providers.base_provider import EmbeddingsBaseProvider
from pydantic import model_validator
from groq import Groq, AsyncGroq
from groq.types import CreateEmbeddingResponse
from typing import List
from typing_extensions import Self

class GroqEmbeddings(EmbeddingsBaseProvider):
    vector_dimensions :int=1024

    @model_validator(mode="after")
    def set_groq(self)->Self:

        self.client :Groq = Groq(
            api_key=self.config.api_key
        )

        self.aclient :AsyncGroq = AsyncGroq(
            api_key=self.config.api_key
        )

        return self
    
    def generate(self, text_batches :List[str])->CreateEmbeddingResponse:
        vectors = self.client.embeddings.create(
            model=self.config.model,
            input=text_batches
        )

        #TODO create base embedding basemodel to map from Mistral EmbeddingResponse

        return vectors
    
    async def agenerate(self, text_batches :List[str])->CreateEmbeddingResponse:
        vectors = self.aclient.embeddings.create(
            model=self.config.model,
            input=text_batches
        )

        return vectors


from aicore.llm.mcp.models import ToolCallSchema, ToolCalls, ToolSchema
from aicore.llm.providers.base_provider import LlmBaseProvider
from aicore.logger import default_stream_handler
from aicore.const import STREAM_START_TOKEN, STREAM_END_TOKEN, REASONING_STOP_TOKEN, TOOL_CALL_START_TOKEN
from pydantic import model_validator
# from mistral_common.protocol.instruct.messages import UserMessage
# from mistral_common.tokens.tokenizers.mistral import MistralTokenizer
from mistralai import Mistral, CompletionEvent, CompletionResponseStreamChoice, models
from typing import Any, Optional, Union, List, Literal, Dict
from typing_extensions import Self
import tiktoken

#TODO replace Tiktoken with Mistral tekken encoder when it is updated to work on python 3.13#
class MistralLlm(LlmBaseProvider):

    @model_validator(mode="after")
    def set_mistral(self)->Self:

        self.client :Mistral = Mistral(
            api_key=self.config.api_key
        )
        self._auth_exception = models.SDKError
        self.validate_config()
        ### Suspect Misral will always stream by default
        self.completion_fn = self.client.chat.stream
        self.acompletion_fn = self.client.chat.stream_async
        self.normalize_fn = self.normalize
        self.tokenizer_fn = tiktoken.encoding_for_model(
            self.get_default_tokenizer(
                self.config.model
            )
        ).encode

        return self
    
    def normalize(self, chunk:CompletionEvent, completion_id :Optional[str]=None)->CompletionResponseStreamChoice:
        data = chunk.data
        # print(f"{data=}")
        if data.usage is not None:
            self.usage.record_completion(
                prompt_tokens=data.usage.prompt_tokens,
                response_tokens=data.usage.completion_tokens,
                completion_id=completion_id or data.id
            )
        return data.choices
    
    def _message_body(self, prompt: Union[List[str], str], role: Literal["user", "system", "assistant"] = "user", img_b64_str: Optional[List[str]] = None, _last: Optional[bool] = False) -> Dict:
        message_body = {
            "role": role,
            "content": self._message_content(prompt, img_b64_str)
        }
        if role == "assistant" and _last:
            message_body["prefix"] = True
        return message_body
    
    def _stream(self, stream, prefix_prompt :Optional[Union[str, List[str]]]=None)->str:
        message = []

        prefix_prompt = "".join(prefix_prompt) if isinstance(prefix_prompt, list) else prefix_prompt
        prefix_buffer = []
        prefix_completed = not bool(prefix_prompt)
        for chunk in stream:
            _chunk = self.normalize_fn(chunk)
            if _chunk:
                chunk_message = _chunk[0].delta.content or ""
                if prefix_completed:
                    default_stream_handler(chunk_message)
                    message.append(chunk_message)
                else:
                    prefix_buffer.append(chunk_message)
                    if "".join(prefix_buffer) == prefix_prompt:
                        prefix_completed = True
                
        if self._is_reasoner:
            default_stream_handler(REASONING_STOP_TOKEN)
        response = "".join(message)
        return response
    
    async def _astream(self, stream, logger_fn, prefix_prompt :Optional[Union[str, List[str]]]=None)->str:
        message = []
    
        await logger_fn(STREAM_START_TOKEN) if not prefix_prompt else ...

        _calling_tool = False
        tool_calls = ToolCalls()

        prefix_prompt = "".join(prefix_prompt) if isinstance(prefix_prompt, list) else prefix_prompt
        prefix_buffer = []
        prefix_completed = not bool(prefix_prompt)
        
        async for chunk in stream:
            _chunk = self.normalize_fn(chunk)
            if _chunk:
                chunk_message = _chunk[0].delta.content or ""
                if prefix_completed and isinstance(chunk_message, str):
                    await logger_fn(chunk_message)
                    message.append(chunk_message)
                elif isinstance(chunk_message, list):
                    ### ignore aditional mistral information like citations for now
                    pass
                else:
                    prefix_buffer.append(chunk_message)
                    if "".join(prefix_buffer) == prefix_prompt:
                        prefix_completed = True
                        await logger_fn(STREAM_START_TOKEN)

            if self._is_tool_call(_chunk):
                ### TODO recheck this line to ensure it covers multiple tool calling in stream mode
                tool_chunk = self._tool_chunk_from_provider(_chunk)
                if not _calling_tool:
                    _calling_tool = True
                    await logger_fn(TOOL_CALL_START_TOKEN)
                    tool_call = self._fill_tool_schema(tool_chunk)
                    continue

                if self._tool_call_change_condition(tool_chunk):
                    if message:
                        ### cover anthropic
                        tool_call._raw = "\n".join(message)
                    tool_calls.root.append(tool_call)
                    tool_call = self._fill_tool_schema(tool_chunk)
                    continue
                
                tool_call = self._handle_tool_call_stream(tool_call, tool_chunk)
        
        if _calling_tool:
            ### colect last call
            if message:
                ### cover anthropic
                tool_call._raw = "\n".join(message)
            tool_calls.root.append(tool_call)
            return tool_calls
        
        if self._is_reasoner:
            await logger_fn(REASONING_STOP_TOKEN)
        else:
            await logger_fn(STREAM_END_TOKEN)
        response = "".join(message)
        return response
    
    @staticmethod
    def _to_provider_tool_schema(tool: ToolSchema) -> Dict[str, Any]:
        """
        Convert to OpenAi tool schema format.
        
        Returns:
            Dictionary in OpenAi tool schema format
        """
        return {
            "type": "function",
            "function": {
                "name": tool.name,
                "description": tool.description,
                "parameters": {
                    "type": tool.input_schema.type,
                    "properties": tool.input_schema.properties.model_dump(),
                    "required": tool.input_schema.required,
                    **{k: v for k, v in tool.input_schema.model_dump().items() 
                       if k not in ["type", "properties", "required"]}
                }
            }
        }
    
    @staticmethod
    def _to_provider_tool_call_schema(toolCallSchema :ToolCallSchema)->ToolCallSchema:
        toolCallSchema._raw = {
            "role": "assistant",
            "tool_calls": [
                {
                    "id": toolCallSchema.id,
                    "function": {
                        "name": toolCallSchema.name,
                        "arguments": toolCallSchema.arguments
                    },
                    "type": "function"
                }
            ]
        }

        return toolCallSchema
    
    def _tool_call_message(self, toolCallSchema :ToolCallSchema, content :str) -> Dict[str, str]:
        return {
            "type": "function_call_output",
            "role": "tool",
            "tool_call_id": toolCallSchema.id,
            "content": str(content)
        }


from typing import Optional

class AiCoreBaseException(Exception):
    def __init__(self, provider :str, message :str, status_code :Optional[int]=401):
        self.provider = provider
        self.message = message
        self.status_code = status_code

    def __str__(self)->str:
        return str(self.__dict__)

class AuthenticationError(AiCoreBaseException):
    ...

class ModelError(AiCoreBaseException):
    def __init__(self, provider :str, message :str, supported_models :Optional[list[str]]=None, status_code :Optional[int]=401):
        super().__init__(provider, message, status_code)
        self.supported_models= supported_models

    @classmethod
    def from_model(cls, model :str, provider :str, supported_models :Optional[list[str]]=None, status_code :Optional[int]=401)->"ModelError":
        return cls(
            provider=provider,
            message=f"Invalid model: {model}",
            supported_models=supported_models,
            status_code=status_code
        )
    
class BalanceError(AiCoreBaseException):
    ...

class FastMcpError(Exception):
    """Exception raised for errors in the FastMcp module.
    
    This exception can wrap other exceptions to maintain their details
    while identifying them as coming from the FastMcp module.
    """
    def __init__(self, message=None, original_exception=None):
        if original_exception and not message:
            # Use the original exception's message if no message is provided
            message = str(original_exception)
        
        super().__init__(message)
        self.original_exception = original_exception

    def __str__(self):
        if self.original_exception:
            return f"FastMcpError: {super().__str__()} (Original exception: {self.original_exception.__class__.__name__})"
        return f"FastMcpError: {super().__str__()}"


from aicore.embeddings.providers.gemini import GeminiEmbeddings
from aicore.embeddings.providers.groq import GroqEmbeddings
from aicore.embeddings.providers.mistral import MistralEmbeddings
from aicore.embeddings.providers.nvidia import NvidiaEmbeddings
from aicore.embeddings.providers.openai import OpenAiEmbeddings
from aicore.embeddings.providers.base_provider import EmbeddingsBaseProvider

__all__ = [
    "OpenAiEmbeddings",
    "MistralEmbeddings",
    "NvidiaEmbeddings",
    "GroqEmbeddings",
    "GeminiEmbeddings",
    "EmbeddingsBaseProvider"
]


import pytest
import time
import requests
import asyncio
from unittest.mock import patch, MagicMock

# Import the decorators from your module
from aicore.utils import (
    retry_on_failure,
    raise_on_balance_error,
    DEFAULT_MAX_ATTEMPTS,
    BalanceError,
    is_out_of_balance,
    should_retry
)

def create_http_error(status_code=429, retry_after=None, text=""):
    """Create a fake requests.HTTPError with specified status code."""
    response = requests.models.Response()
    response.status_code = status_code
    response._content = text.encode() if text else b''
    if retry_after is not None:
        response.headers['Retry-After'] = str(retry_after)
    error = requests.exceptions.HTTPError(f"{status_code} Error", response=response)
    return error

class CustomException(Exception):
    """Generic custom exception for testing"""
    pass

class BalanceErrorSimulator(Exception):
    """Simulates balance error with 400 status and credit message"""
    def __init__(self):
        self.response = MagicMock()
        self.response.status_code = 400
        self.response.text = "credit balance is too low"
        self.response.json.return_value = {"error": {"message": "credit balance is too low"}}
        super().__init__("credit balance is too low")

    def __str__(self):
        return f"{self.response.status_code} {self.response.text}"

# Tests for helper functions
def test_should_retry():
    """Test the should_retry function logic"""
    # Should retry normal exceptions
    assert should_retry(Exception("Normal error")) is True
    assert should_retry(CustomException("Custom error")) is True
    
    # Should not retry 400 errors
    assert should_retry(create_http_error(400)) is False
    
    # Should retry other HTTP errors
    assert should_retry(create_http_error(429)) is True
    assert should_retry(create_http_error(500)) is True
    
    # Should not retry balance errors
    assert should_retry(BalanceErrorSimulator()) is False

def test_is_out_of_balance():
    """Test the is_out_of_balance function logic"""
    # Should detect balance errors
    assert is_out_of_balance(BalanceErrorSimulator()) is True
    
    # Should detect HTTP errors with balance messages
    balance_error = create_http_error(400, text='{"error": {"message": "credit balance is too low"}')
    balance_error.response.json = MagicMock(return_value={"error": {"message": "credit balance is too low"}})
    assert is_out_of_balance(balance_error) is True
    
    # Should not flag non-balance errors
    assert is_out_of_balance(Exception("Normal error")) is False
    assert is_out_of_balance(create_http_error(429)) is False

# Sync Tests
def test_retry_on_generic_error_sync(monkeypatch):
    """Test that generic errors are retried in sync functions"""
    call_count = 0
    sleep_calls = []

    monkeypatch.setattr(time, "sleep", lambda t: sleep_calls.append(t))

    @retry_on_failure
    def sometimes_fail():
        nonlocal call_count
        call_count += 1
        if call_count < 3:
            raise CustomException("Temporary failure")
        return "success"

    result = sometimes_fail()
    assert result == "success"
    assert call_count == 3
    assert len(sleep_calls) == 2  # Should have slept between attempts

def test_max_retries_reached_sync(monkeypatch):
    """Test that max retries are respected in sync functions"""
    call_count = 0
    sleep_calls = []

    monkeypatch.setattr(time, "sleep", lambda t: sleep_calls.append(t))

    @retry_on_failure
    def always_fail():
        nonlocal call_count
        call_count += 1
        raise CustomException("Persistent failure")

    result = always_fail()
    assert result is None
    assert call_count == DEFAULT_MAX_ATTEMPTS
    assert len(sleep_calls) == DEFAULT_MAX_ATTEMPTS - 1  # Should sleep between each attempt

def test_no_retry_on_400_sync():
    """Test that 400 errors are not retried in sync functions"""
    call_count = 0

    @retry_on_failure
    def fail_with_400():
        nonlocal call_count
        call_count += 1
        raise create_http_error(400)

    result = fail_with_400()
    assert result is None
    assert call_count == 1  # Should only call once, no retries

def test_balance_error_conversion_sync():
    """Test that balance errors are properly converted in sync functions"""
    call_count = 0

    @raise_on_balance_error
    @retry_on_failure
    def fail_with_balance_error():
        nonlocal call_count
        call_count += 1
        raise BalanceErrorSimulator()

    with pytest.raises(BalanceError) as exc_info:
        fail_with_balance_error()
    
    assert "credit balance is too low" in str(exc_info.value)
    assert call_count == 1  # Should only call once, no retries for balance errors

def test_retry_with_http_429_sync(monkeypatch):
    """Test retry behavior with HTTP 429 (rate limit) errors"""
    call_count = 0
    sleep_calls = []

    monkeypatch.setattr(time, "sleep", lambda t: sleep_calls.append(t))

    @retry_on_failure
    def fail_with_rate_limit():
        nonlocal call_count
        call_count += 1
        if call_count < 3:
            raise create_http_error(429, retry_after="2")
        return "success"

    result = fail_with_rate_limit()
    assert result == "success"
    assert call_count == 3
    assert len(sleep_calls) >= 2  # Should have slept between attempts

# # Async Tests
@pytest.mark.asyncio
async def test_retry_on_generic_error_async(monkeypatch):
    """Test that generic errors are retried in async functions"""
    call_count = 0
    sleep_calls = []

    # Mock asyncio.sleep
    async def mock_sleep(t):
        sleep_calls.append(t)
        return

    monkeypatch.setattr(asyncio, "sleep", mock_sleep)

    @retry_on_failure
    async def sometimes_fail():
        nonlocal call_count
        call_count += 1
        if call_count < 3:
            raise CustomException("Temporary failure")
        return "success"

    result = await sometimes_fail()
    assert result == "success"
    assert call_count == 3
    assert len(sleep_calls) == 2  # Should have slept between attempts

@pytest.mark.asyncio
async def test_max_retries_reached_async(monkeypatch):
    """Test that max retries are respected in async functions"""
    call_count = 0
    sleep_calls = []

    # Mock asyncio.sleep
    async def mock_sleep(t):
        sleep_calls.append(t)
        return

    monkeypatch.setattr(asyncio, "sleep", mock_sleep)

    @retry_on_failure
    async def always_fail():
        nonlocal call_count
        call_count += 1
        raise CustomException("Persistent failure")

    result = await always_fail()
    assert result is None
    assert call_count == DEFAULT_MAX_ATTEMPTS
    assert len(sleep_calls) == DEFAULT_MAX_ATTEMPTS - 1  # Should sleep between each attempt

@pytest.mark.asyncio
async def test_no_retry_on_400_async():
    """Test that 400 errors are not retried in async functions"""
    call_count = 0

    @retry_on_failure
    async def fail_with_400():
        nonlocal call_count
        call_count += 1
        raise create_http_error(400)

    result = await fail_with_400()
    assert result is None
    assert call_count == 1  # Should only call once, no retries

@pytest.mark.asyncio
async def test_balance_error_conversion_async():
    """Test that balance errors are properly converted in async functions"""
    call_count = 0

    @raise_on_balance_error
    @retry_on_failure
    async def fail_with_balance_error():
        nonlocal call_count
        call_count += 1
        raise BalanceErrorSimulator()

    with pytest.raises(BalanceError) as exc_info:
        await fail_with_balance_error()
    
    assert "credit balance is too low" in str(exc_info.value)
    assert call_count == 1  # Should only call once, no retries for balance errors

@pytest.mark.asyncio
async def test_retry_with_http_429_async(monkeypatch):
    """Test retry behavior with HTTP 429 (rate limit) errors in async functions"""
    call_count = 0
    sleep_calls = []

    # Mock asyncio.sleep
    async def mock_sleep(t):
        sleep_calls.append(t)
        return

    monkeypatch.setattr(asyncio, "sleep", mock_sleep)

    @retry_on_failure
    async def fail_with_rate_limit():
        nonlocal call_count
        call_count += 1
        if call_count < 3:
            raise create_http_error(429, retry_after="2")
        return "success"

    result = await fail_with_rate_limit()
    assert result == "success"
    assert call_count == 3
    assert len(sleep_calls) == 2  # Should have slept between attempts

def test_decorator_order_handling():
    """Test that decorator order is correct (retry first, then balance check)"""
    call_count = 0

    @raise_on_balance_error
    @retry_on_failure
    def mixed_failure():
        nonlocal call_count
        call_count += 1
        if call_count < 3:
            raise CustomException("Temporary failure")
        raise BalanceError(provider="dummy", message="str")

    with pytest.raises(BalanceError):  # More specific exception
        mixed_failure()
    
    assert call_count == 3  # Should have retried before hitting balance error

@pytest.mark.asyncio
async def test_decorator_order_handling_async():
    """Test that decorator order is correct for async functions"""
    call_count = 0

    @raise_on_balance_error
    @retry_on_failure
    async def mixed_failure():
        nonlocal call_count
        call_count += 1
        if call_count < 3:
            raise CustomException("Temporary failure")
        raise BalanceError(provider="dummy", message="str")

    with pytest.raises(BalanceError):  # More specific exception
        await mixed_failure()
    
    assert call_count == 3  # Should have retried before hitting balance error

# def test_returning_none_doesnt_trigger_error():
#     """Test that functions that return None don't trigger errors"""
#     call_count = 0

#     @retry_on_failure
#     def return_none():
#         nonlocal call_count
#         call_count += 1
#         return None

#     result = return_none()
#     assert result is None
#     assert call_count == 1  # Should only call once, no errors

# def test_retry_on_specific_error_types():
#     """Test retry behavior with different error types"""
#     results = []

#     @retry_on_failure
#     def raise_different_errors():
#         if not results:
#             results.append("attempt1")
#             raise TimeoutError("Connection timeout")
#         elif len(results) == 1:
#             results.append("attempt2")
#             raise ConnectionError("Network error")
#         else:
#             results.append("success")
#             return "success"

#     result = raise_different_errors()
#     assert result == "success"
#     assert len(results) == 3
#     assert "attempt1" in results
#     assert "attempt2" in results
#     assert "success" in results


from pydantic import BaseModel, model_validator
from datetime import datetime, timedelta
from typing import Optional, Dict
import pytz
import json

from aicore.const import METADATA_JSON, DEFAULT_ENCODING

with open(METADATA_JSON, "r", encoding=DEFAULT_ENCODING) as _file:    
    ### https://www.anthropic.com/pricing#anthropic-api
    ### https://openai.com/api/pricing/    
    ### https://mistral.ai/products/la-plateforme#pricing
    ### https://groq.com/pricing/    
    ### https://api-docs.deepseek.com/quick_start/pricing
    MODELS_METADATA: Dict = json.load(_file)

class HappyHour(BaseModel):
    start: datetime
    finish: datetime
    pricing: "PricingConfig"

    @model_validator(mode="before")
    @classmethod
    def parse_time_strings(cls, kwargs: dict) -> dict:
        parsed_args = {}
        for key, value in kwargs.items():
            if key == "pricing":
                parsed_args[key] = value
                continue

            if isinstance(value, datetime):
                # If already a datetime, ensure it's pytz.UTC
                if value.tzinfo is None:
                    parsed_args[key] = value.replace(tzinfo=pytz.UTC)
                    continue
                parsed_args[key] = value.astimezone(pytz.UTC)
                
            elif isinstance(value, str):
                try:
                    # Parse time string (e.g. "16:30")
                    time_obj = datetime.strptime(value, "%H:%M").time()
                    # Get today's date in pytz.UTC
                    today = datetime.now(pytz.UTC).date()
                    # Combine date and time
                    naive_dt = datetime.combine(today, time_obj)
                    # Handle overnight case (e.g. finish time is next day)
                    if key == 'finish' and time_obj.hour < 12:
                        naive_dt += timedelta(days=1)
                    # Make timezone aware
                    parsed_args[key] = naive_dt.replace(tzinfo=pytz.UTC)
                    
                except ValueError as e:
                    raise ValueError(f"Invalid time format: {value}. Expected HH:MM") from e
                        
        return parsed_args

class DynamicPricing(BaseModel):
    threshold: int
    pricing: "PricingConfig"

    @model_validator(mode="after")
    def validate_threshold(self) -> "DynamicPricing":
        if self.threshold <= 0:
            raise ValueError("Dynamic pricing threshold must be positive")
        return self

class PricingConfig(BaseModel):
    """
    pricing ($) per 1M tokens
    """
    input: float
    output: float = 0
    cached: float = 0
    cache_write: float = 0
    happy_hour: Optional[HappyHour] = None
    dynamic: Optional[DynamicPricing] = None
    
    def calculate_cost(self, 
            prompt_tokens: int, 
            response_tokens: int, 
            cached_tokens: int = 0, 
            cache_write_tokens: int = 0,
            timestamp: Optional[datetime] = None
        ) -> float:
        """Calculate cost based on token counts and current pricing rules"""
        pricing = self._get_active_pricing(timestamp)
        
        cost = (pricing.input * prompt_tokens + 
                pricing.output * response_tokens + 
                pricing.cached * cached_tokens + 
                pricing.cache_write * cache_write_tokens)
        
        return cost * 1e-6  # Convert from per 1M tokens to per token

    def _get_active_pricing(self, timestamp: Optional[datetime] = None) -> "PricingConfig":
        timestamp = timestamp or datetime.now(pytz.UTC)
        if self.happy_hour and self.happy_hour.start <= timestamp <= self.happy_hour.finish:
            return self.happy_hour.pricing
        return self

class ModelMetaData(BaseModel):
    context_window: int = 128000
    max_tokens: int = 8192
    tool_use: bool = True
    pricing: Optional[PricingConfig] = None

METADATA: Dict[str, ModelMetaData] = {
    model: ModelMetaData(**metadata)
    for model, metadata in MODELS_METADATA.items()
}




"""
AI Core: A unified interface for LLM and embedding providers.

This library provides a consistent API for working with various LLM providers
(OpenAI, Mistral, Groq, Gemini, Nvidia, OpenRouter) and embedding providers,
with support for both synchronous and asynchronous operations.
"""

from aicore.config import Config
from aicore.llm import Llm, LlmConfig
from aicore.embeddings import Embeddings, EmbeddingsConfig
from aicore.logger import Logger, _logger

__all__ = [
    "Config",
    "Llm",
    "LlmConfig",
    "Embeddings",
    "EmbeddingsConfig",
    "Logger"
]

__version__ = "0.1.9"


import os
import ulid
import json
import pytest
import tempfile
from unittest.mock import patch, MagicMock, mock_open
import polars as pl
from pathlib import Path
from sqlalchemy import create_engine
from sqlalchemy.orm import sessionmaker

from aicore.const import DEFAULT_OBSERVABILITY_DIR, DEFAULT_OBSERVABILITY_FILE, DEFAULT_ENCODING
# Use the collector module that now also supports PostgreSQL.
from aicore.observability.collector import LlmOperationCollector, LlmOperationRecord
from aicore.observability.models import Session, Message, Metric

class FakeCursor:
    """A fake database cursor that tracks executed SQL commands."""
    def __init__(self, executed_queries: list):
        self.executed_queries = executed_queries

    def execute(self, query, params=None):
        self.executed_queries.append(query)

    def close(self):
        pass

class FakeConn:
    """A fake PostgreSQL connection that provides a fake cursor and commit behaviour."""
    def __init__(self, executed_queries: list = None):
        if executed_queries is None:
            executed_queries = []
        self.executed_queries = executed_queries

    def cursor(self):
        return FakeCursor(self.executed_queries)

    def commit(self):
        pass

class TestLlmOperationRecord:
    def test_init_with_minimal_args(self):
        """Test initializing with only required fields."""
        record = LlmOperationRecord(
            operation_type="completion",
            provider="openai",
            completion_args={"model": "gpt-4"},
            latency_ms=100.0
        )
        assert record.operation_type == "completion"
        assert record.provider == "openai"
        assert record.latency_ms == 100.0
        assert record.completion_args == {"model": "gpt-4"}
        assert record.timestamp is not None
        assert record.operation_id is not None

    def test_field_validators(self):
        """Test field validators work correctly."""
        # Test timestamp validator
        record = LlmOperationRecord(
            operation_type="completion",
            provider="openai",
            completion_args={"model": "gpt-4"},
            latency_ms=100.0
        )
        assert isinstance(record.timestamp, str)

        # Test completion_args validator with string
        json_args = json.dumps({"model": "gpt-4", "temperature": 0.7})
        record = LlmOperationRecord(
            operation_type="completion",
            provider="openai",
            completion_args=json_args,
            latency_ms=100.0
        )
        assert isinstance(record.completion_args, dict)
        assert record.completion_args["model"] == "gpt-4"

        # Test response validator with dict
        response_dict = {"choices": [{"message": {"content": "Hello"}}]}
        record = LlmOperationRecord(
            operation_type="completion",
            provider="openai",
            completion_args={"model": "gpt-4"},
            latency_ms=100.0,
            response=response_dict
        )
        assert isinstance(record.response, str)
        assert "Hello" in record.response

    def test_computed_fields(self):
        """Test computed fields return correct values."""
        record = LlmOperationRecord(
            operation_type="completion",
            provider="openai",
            completion_args={
                "model": "gpt-4",
                "temperature": 0.7,
                "max_tokens": 1000,
                "messages": [
                    {"role": "system", "content": "You are a helpful assistant"},
                    {"role": "user", "content": "Hello"},
                    {"role": "assistant", "content": "Hi there"}
                ]
            },
            input_tokens=10,
            output_tokens=5,
            latency_ms=100.0
        )
        assert record.model == "gpt-4"
        assert record.temperature == 0.7
        assert record.max_tokens == 1000
        assert record.system_prompt == "You are a helpful assistant"
        assert record.user_prompt == "Hello"
        assert record.assistant_message == "Hi there"
        assert record.total_tokens == 15
        assert record.success is False  # Since response is None

        # Test with response
        record.response = "Test response"
        assert record.success is True

class TestLlmOperationCollectorFileStorage:
    def test_init(self):
        """Test initialization of the collector."""
        collector = LlmOperationCollector()
        assert collector.root == []
        assert collector.storage_path is None

    def test_storage_path_setter(self):
        """Test setting storage path."""
        collector = LlmOperationCollector()
        collector.storage_path = "test/path.json"
        assert collector.storage_path == "test/path.json"
        
        # Test with Path object
        path_obj = Path("/tmp/test.json")
        collector.storage_path = path_obj
        assert collector.storage_path == path_obj

    def test_record_completion(self):
        """Test recording a completion operation."""
        collector = LlmOperationCollector()
        record = collector.record_completion(
            completion_args={"model": "gpt-4", "messages": [{"role": "user", "content": "Hello"}]},
            operation_type="completion",
            provider="openai",
            response="Test response",
            input_tokens=10,
            output_tokens=5,
            latency_ms=150.0
        )
        assert len(collector.root) == 1
        assert collector.root[0] == record
        assert record.provider == "openai"
        assert record.operation_type == "completion"
        assert record.response == "Test response"
        assert record.input_tokens == 10
        assert record.output_tokens == 5
        assert record.latency_ms == 150.0

    def test_clean_completion_args(self):
        """Test cleaning of sensitive information from completion arguments."""
        collector = LlmOperationCollector()
        args = {
            "model": "gpt-4",
            "api_key": "sk-sensitive-key-12345",
            "messages": [{"role": "user", "content": "Hello"}]
        }
        cleaned = collector._clean_completion_args(args)
        assert "api_key" not in cleaned
        assert cleaned["model"] == "gpt-4"
        assert cleaned["messages"] == [{"role": "user", "content": "Hello"}]

    @patch.dict('os.environ', {"OBSERVABILITY_DATA_DEFAULT_FILE": "/env/path.json"})
    def test_from_observable_storage_path_with_env(self):
        """Test creating collector with environment variable path."""
        collector = LlmOperationCollector.fom_observable_storage_path()
        assert collector.storage_path == "/env/path.json"

    def test_from_observable_storage_path_with_param(self):
        """Test creating collector with provided path."""
        collector = LlmOperationCollector.fom_observable_storage_path("/custom/path.json")
        assert collector.storage_path == "/custom/path.json"

    @patch('pathlib.Path.__truediv__', return_value=Path("/default/path/llm_operations.json"))
    def test_from_observable_storage_path_default(self, mock_truediv):
        """Test creating collector with default path."""
        with patch.dict('os.environ', {}, clear=True):
            collector = LlmOperationCollector.fom_observable_storage_path()
            expected_path = Path(DEFAULT_OBSERVABILITY_DIR) / DEFAULT_OBSERVABILITY_FILE
            assert collector.storage_path == expected_path


class TestIntegrationFileStorage:
    """Integration tests for the collector and record classes (file storage)."""
    
    def test_end_to_end(self):
        """Test an end-to-end flow with a temporary file."""
        with tempfile.TemporaryDirectory() as temp_dir:
            temp_file = os.path.join(temp_dir, "test_records.json")
            
            # Create collector with storage path
            collector = LlmOperationCollector()
            collector.storage_path = temp_file
            
            # Record a completion
            collector.record_completion(
                completion_args={
                    "model": "gpt-4",
                    "messages": [{"role": "user", "content": "Hello"}],
                    "temperature": 0.5
                },
                operation_type="completion",
                provider="openai",
                response="Hi there!",
                input_tokens=5,
                output_tokens=3,
                latency_ms=120.0
            )
            
            # Verify the file was created and contains the record
            assert os.path.exists(temp_file)
            
            with open(temp_file, 'r', encoding='utf-8') as f:
                data = json.load(f)
                
            # Expect a list of serialized records (the structure remains as before)
            assert len(data) == 1
            record = data[0]
            assert record["provider"] == "openai"
            assert record["model"] == "gpt-4"
            assert record["response"] == "Hi there!"
            assert record["temperature"] == 0.5
            
            # Add another record
            collector.record_completion(
                completion_args={
                    "model": "gpt-3.5-turbo",
                    "messages": [{"role": "user", "content": "How are you?"}]
                },
                operation_type="completion",
                provider="openai",
                response="I'm fine, thank you!",
                input_tokens=4,
                output_tokens=5,
                latency_ms=80.0
            )
            
            # Verify both records are in the file
            with open(temp_file, 'r', encoding='utf-8') as f:
                data = json.load(f)
                
            assert len(data) == 2
            assert data[1]["model"] == "gpt-3.5-turbo"
            assert data[1]["response"] == "I'm fine, thank you!"


# --- Integration Storage Tests (File) ---
class TestIntegrationStorage:
    """
    Integration tests for LlmOperationCollector and LlmOperationRecord, including file storage.
    """
    def test_end_to_end_storage(self):
        """
        Test an end-to-end flow where a record is stored in a temporary file.
        """
        with tempfile.TemporaryDirectory() as temp_dir:
            temp_file = os.path.join(temp_dir, "test_records.json")
            
            collector = LlmOperationCollector()
            collector.storage_path = temp_file
            
            collector.record_completion(
                completion_args={
                    "model": "gpt-4",
                    "messages": [{"role": "user", "content": "Hello"}],
                    "temperature": 0.5
                },
                operation_type="completion",
                provider="openai",
                response="Hi there!",
                input_tokens=5,
                output_tokens=3,
                latency_ms=120.0
            )
            
            assert os.path.exists(temp_file), "Storage file was not created."
            
            with open(temp_file, 'r', encoding='utf-8') as f:
                data = json.load(f)
            
            assert len(data) >= 1, "Expected at least one record in the storage file."
            record = data[0]
            assert record["provider"] == "openai"
            assert record["model"] == "gpt-4"
            assert record["response"] == "Hi there!"
            assert record["temperature"] == 0.5


@pytest.fixture
def in_memory_collector(monkeypatch, tmp_path):
    """
    Create a collector with an in-memory SQLite DB and temporary file storage.
    """
    # Use in-memory SQLite for testing DB operations that occur within the same collector instance.
    monkeypatch.setenv("CONNECTION_STRING", "sqlite:///:memory:")
    collector = LlmOperationCollector().init_dbsession()
    storage_file = tmp_path / "observability.json"
    collector.storage_path = str(storage_file)
    return collector


@pytest.fixture
def file_based_db_collector(monkeypatch, tmp_path):
    """
    Create a collector using a file-based SQLite database. This is necessary for tests that use
    class methods (e.g. polars_from_db, get_filter_options, get_metrics_summary) since these methods
    create a new instance. Using a file-based DB ensures state is shared.
    """
    db_file = tmp_path / "test.db"
    db_url = "sqlite:///" + str(db_file)
    monkeypatch.setenv("CONNECTION_STRING", db_url)
    collector = LlmOperationCollector().init_dbsession()
    # Use a temporary file for file-storage even if not needed by DB methods.
    storage_file = tmp_path / "observability.json"
    collector.storage_path = str(storage_file)
    return collector


# For tests that call class methods (which create a new instance) we patch __init__
# so that the new instance is set up with the same engine/DBSession.
def patch_init_with_db(monkeypatch, db_url: str):
    engine = create_engine(db_url)
    SessionMaker = sessionmaker(bind=engine)
    monkeypatch.setattr(
        LlmOperationCollector,
        "init_dbsession",
        lambda self: (setattr(self, "_dbsession", SessionMaker), setattr(self, "engine", engine))
    )


# === Tests =========================================================

def test_init_dbsession(monkeypatch):
    """Test that init_dbsession properly sets up the DB connection and engine."""
    monkeypatch.setenv("CONNECTION_STRING", "sqlite:///:memory:")
    collector = LlmOperationCollector().init_dbsession()
    assert collector._table_initialized is True
    assert collector._engine is not None
    assert collector._session_factory is not None


def test_clean_completion_args():
    """Test that _clean_completion_args removes the 'api_key' key."""
    args = {"param": "value", "api_key": "secret123"}
    cleaned = LlmOperationCollector._clean_completion_args(args)
    assert "api_key" not in cleaned
    assert cleaned.get("param") == "value"


def test_record_completion_file_storage(in_memory_collector):
    """
    Test that record_completion:
      - Cleans the arguments,
      - Appends the new record to the root,
      - Writes the record to the file (using JSON format).
    """
    args = {"param": "value", "api_key": "should_be_removed"}
    record = in_memory_collector.record_completion(
        completion_args=args,
        operation_type="completion",
        provider="test_provider",
        response="test response",
        session_id="session1",
        workspace="workspace1",
        agent_id="agent1",
        action_id="action1",
        input_tokens=10,
        output_tokens=5,
        cost=0.05,
        latency_ms=100,
        error_message=""
    )
    # Verify the record was appended in memory.
    assert record in in_memory_collector.root

    # Verify the file storage. (The _store_to_file method writes a JSON array.)
    with open(in_memory_collector.storage_path, "r", encoding=DEFAULT_ENCODING) as f:
        data = json.loads(f.read())
    # Look for our record by its operation_id.
    found = any(item.get("operation_id") == record.operation_id for item in data)
    assert found

    # Also ensure the sensitive key was removed in the stored JSON.
    for item in data:
        if item.get("operation_id") == record.operation_id:
            # completion_args was JSON-dumped, so load it back to check.
            completion_args = json.loads(item["completion_args"])
            assert "api_key" not in completion_args


def test_record_completion_db(monkeypatch, tmp_path):
    """
    Test that record_completion inserts records into the database.
    (Here we use an in-memory DB so that the same collector instance is used.)
    """
    monkeypatch.setenv("CONNECTION_STRING", "sqlite:///:memory:")
    collector = LlmOperationCollector().init_dbsession()
    storage_file = tmp_path / "observability.json"
    collector.storage_path = str(storage_file)

    args = {"param": "value", "api_key": "secret123"}
    record = collector.record_completion(
        completion_args=args,
        operation_type="completion",
        provider="test_provider",
        response="test response db",
        session_id="session_db",
        workspace="workspace_db",
        agent_id="agent_db",
        action_id="action_db",
        input_tokens=15,
        output_tokens=7,
        cost=0.10,
        latency_ms=150,
        error_message=""
    )
    # _last_inserted_record should match.
    assert collector._last_inserted_record == record.operation_id

    # Use the collector’s DBSession to query the inserted records.
    session_local = collector._session_factory()
    db_sess = session_local.query(Session).filter_by(session_id=record.session_id).first()
    assert db_sess is not None

    db_message = session_local.query(Message).filter_by(operation_id=record.operation_id).first()
    assert db_message is not None

    db_metric = session_local.query(Metric).filter_by(operation_id=record.operation_id).first()
    assert db_metric is not None

    session_local.close()


def test_polars_from_file(file_based_db_collector, tmp_path):
    """
    Test that polars_from_file returns a non-empty DataFrame when the file contains valid records.
    We simulate file storage by writing a JSON list of a dummy record.
    """
    # Create a dummy record dictionary (mimicking a serialized LlmOperationRecord)
    dummy_record = {
        "session_id": "sess1",
        "workspace": "ws1",
        "agent_id": "agent1",
        "action_id": "act1",
        "timestamp": "2025-03-18T00:00:00Z",
        "operation_id": str(ulid.ulid()),
        "operation_type": "completion",
        "provider": "test",
        "model": "dummy",
        "system_prompt": "",
        "user_prompt": "",
        "response": "ok",
        "success": True,
        "assistant_message": "",
        "history_messages": "",
        "temperature": "",
        "max_tokens": 0,
        "input_tokens": 0,
        "output_tokens": 0,
        "total_tokens": 0,
        "cost": 0.0,
        "latency_ms": 100.0,
        "error_message": "",
        "completion_args": json.dumps({"param": "value"}, indent=4)
    }
    # Write the dummy record as a JSON list to the storage file.
    with open(file_based_db_collector.storage_path, "w", encoding=DEFAULT_ENCODING) as f:
        f.write(json.dumps([dummy_record]))
    
    df = LlmOperationCollector.polars_from_file(file_based_db_collector.storage_path)
    assert isinstance(df, pl.DataFrame)
    assert not df.is_empty()
    assert "session_id" in df.columns


def test_polars_from_db(monkeypatch, tmp_path):
    """
    Test that polars_from_db returns a DataFrame that includes the inserted record.
    Here we use a file-based SQLite DB so that a new instance (created by the class method)
    sees the inserted record.
    """
    db_file = tmp_path / "test.db"
    db_url = "sqlite:///" + str(db_file)
    monkeypatch.setenv("CONNECTION_STRING", db_url)
    
    # Use file-based DB collector to insert a record.
    collector = LlmOperationCollector().init_dbsession()
    storage_file = tmp_path / "observability.json"
    collector.storage_path = str(storage_file)
    
    args = {"param": "value", "api_key": "secret123"}
    collector.record_completion(
        completion_args=args,
        operation_type="completion",
        provider="db_provider",
        response="db response",
        session_id="sess_db",
        workspace="ws_db",
        agent_id="agent_db",
        action_id="action_db",
        input_tokens=20,
        output_tokens=10,
        cost=0.20,
        latency_ms=200,
        error_message=""
    )
    # Patch __init__ so that new instances used in the class method have a valid DBSession.
    patch_init_with_db(monkeypatch, db_url)
    
    df = LlmOperationCollector.polars_from_db()
    assert isinstance(df, pl.DataFrame)
    assert not df.is_empty()
    expected_cols = ["session_id", "workspace", "agent_id", "action_id", "operation_id"]
    for col in expected_cols:
        assert col in df.columns


"""
Base provider class for LLM implementations with common functionality.
This module defines the base class for all LLM providers in the system, providing
common interfaces and utilities for synchronous and asynchronous operations.
"""

from aicore.llm.config import LlmConfig
from aicore.llm.mcp.client import MCPClient
from aicore.llm.mcp.models import ToolCallSchema, ToolCalls, ToolSchema
from aicore.logger import _logger, default_stream_handler
from aicore.const import REASONING_START_TOKEN, REASONING_STOP_TOKEN, STREAM_START_TOKEN, STREAM_END_TOKEN, CUSTOM_MODELS, TOOL_CALL_END_TOKEN, TOOL_CALL_START_TOKEN
from aicore.llm.utils import parse_content, image_to_base64
from aicore.llm.usage import UsageInfo
from aicore.models import AuthenticationError, ModelError
from aicore.models_metadata import METADATA
from aicore.observability.collector import LlmOperationCollector
from typing import Any, Dict, Optional, Literal, List, Tuple, Union, Callable
from pydantic import BaseModel, RootModel, Field
from functools import partial, wraps
from json_repair import repair_json
from pathlib import Path
import tiktoken
import asyncio
import json
import time
import ulid

class LlmBaseProvider(BaseModel):    
    """Base class for all LLM provider implementations.
    
    Provides common functionality for:
    - Configuration management
    - Session tracking
    - Synchronous and asynchronous completions
    - Streaming support
    - Observability integration
    - Usage tracking
    
    Attributes:
        config: Configuration for the LLM provider
        session_id: Unique session identifier
        workspace: Optional workspace identifier
        agent_id: Optional agent identifier
        extras: Additional provider-specific data
        _client: Synchronous client instance
        _aclient: Asynchronous client instance
        _collector: Observability collector instance
    """
    config: LlmConfig
    session_id: str = Field(default_factory=ulid.ulid)
    worspace: Optional[str]=None
    agent_id: Optional[str]=None
    extras: Optional[dict]=Field(default_factory=dict)
    tools: Optional[List[ToolSchema]]=None
    _client: Any = None
    _aclient: Any = None
    _completion_args: Dict = {}
    _completion_fn: Any = None
    _acompletion_fn: Any = None
    _normalize_fn: Any = None
    _tokenizer_fn: Any = None
    _is_reasoner: bool = False
    _auth_exception: Exception = Exception
    _usage :Optional[UsageInfo]=None
    _collector: Optional[LlmOperationCollector] = None
    _mcp: Optional[MCPClient] = None
    _n_sucessive_tool_calls :int=0

    @classmethod
    def from_config(cls, config: LlmConfig) -> "LlmBaseProvider":
        """Create provider instance from configuration.
        
        Args:
            config: LLM configuration to initialize the provider
            
        Returns:
            LlmBaseProvider: Initialized provider instance
            
        Example:
            >>> config = LlmConfig(provider="openai", api_key="...", model="gpt-4")
            >>> provider = LlmBaseProvider.from_config(config)
        """
        return cls(
            config=config
        )
    
    @property
    def client(self):
        """Get the synchronous client instance.
        
        Returns:
            The configured synchronous client
        """
        return self._client
    
    @client.setter
    def client(self, client: Any):
        """Set the synchronous client instance.
        
        Args:
            client: Client instance to set
        """
        self._client = client

    @property
    def aclient(self):
        """Get the asynchronous client instance.
        
        Returns:
            The configured asynchronous client
        """
        return self._aclient
    
    @aclient.setter
    def aclient(self, aclient: Any):
        """Set the asynchronous client instance.
        
        Args:
            aclient: Async client instance to set
        """
        self._aclient = aclient

    def validate_config(self, force_check_against_provider :bool=False):
        """Validate provider configuration against available models.
        
        Args:
            exception: Exception type to raise if validation fails
            
        Raises:
            ModelError: If configured model is not available
            AuthenticationError: If provider authentication fails
        """
        try:
            if self.config.model in CUSTOM_MODELS:
                return
            
            if not force_check_against_provider and self.config.provider_model in METADATA:
                return
            
            models = self.client.models.list()
            models = [model.id for model in models.data]
            if self.config.model not in models:
                if self.config.model.endswith("-latest"):
                    model_id = "-".join(self.config.model.split("-")[:-1])
                    models_id = ["-".join(model.split("-")[:-1]) for model in models]
                    if model_id in models_id:
                        return
                elif f"models/{self.config.model}" in models:
                    return
                raise ModelError.from_model(self.config.model, self.config.provider, models)
        except self._auth_exception as e:
            raise AuthenticationError(
                provider=self.config.provider,
                message=str(e)
            )

    @property
    def completion_args(self) -> Dict:
        """Get additional completion arguments.
        
        Returns:
            Dictionary of completion arguments
        """
        return self._completion_args
    
    @completion_args.setter
    def completion_args(self, args: Dict):
        """Set additional completion arguments.
        
        Args:
            args: Dictionary of completion arguments to set
        """
        self._completion_args = args

    @property
    def completion_fn(self) -> Any:
        """Get the synchronous completion function.
        
        Returns:
            The configured completion function
        """
        return self._completion_fn
    
    @completion_fn.setter
    def completion_fn(self, completion_fn: Any):
        """Set the synchronous completion function.
        
        Args:
            completion_fn: Completion function to set
        """
        self._completion_fn = completion_fn

    @property
    def acompletion_fn(self) -> Any:
        """Get the asynchronous completion function.
        
        Returns:
            The configured async completion function
        """
        return self._acompletion_fn
    
    @acompletion_fn.setter
    def acompletion_fn(self, acompletion_fn: Any):
        """Set the asynchronous completion function.
        
        Args:
            acompletion_fn: Async completion function to set
        """
        self._acompletion_fn = acompletion_fn

    @property
    def normalize_fn(self) -> Any:
        """Get the response normalization function.
        
        Returns:
            The configured normalization function
        """
        return self._normalize_fn
    
    @normalize_fn.setter
    def normalize_fn(self, normalize_fn: Any):
        """Set the response normalization function.
        
        Args:
            normalize_fn: Normalization function to set
        """
        self._normalize_fn = normalize_fn

    @property
    def tokenizer_fn(self) -> Any:
        """Get the tokenizer function.
        
        Returns:
            The configured tokenizer function
        """
        return self._tokenizer_fn
    
    @tokenizer_fn.setter
    def tokenizer_fn(self, tokenizer_fn: Any):
        """Set the tokenizer function.
        
        Args:
            tokenizer_fn: Tokenizer function to set
        """
        self._tokenizer_fn = tokenizer_fn

    @property
    def usage(self) -> UsageInfo:
        """Get usage information tracker.
        
        Returns:
            UsageInfo instance tracking token usage and costs
        """
        if self._usage is None:
            self._usage = UsageInfo.from_pricing_config(self.config.pricing)
        return self._usage
    
    @usage.setter
    def usage(self, usage :UsageInfo):
        """Set usage information tracker.
        
        Args:
            usage: UsageInfo instance to set
        """
        self._usage = usage
    
    @property
    def collector(self) -> Optional[LlmOperationCollector]:
        """Get the operation collector instance.
        
        Returns:
            LlmOperationCollector instance if configured, None otherwise
        """
        if self._collector is None:
            self._collector = LlmOperationCollector.fom_observable_storage_path()
        return self._collector
    
    @collector.setter
    def collector(self, collector: LlmOperationCollector):
        """Set the operation collector instance.
        
        Args:
            collector: LlmOperationCollector instance to set
        """
        self._collector = collector
        
    def disable_collection(self):
        """Disable data collection for this provider."""
        if self._collector:
            self._collector.is_enabled = False

    @property
    def mcp(self)->MCPClient:
        if self.config.mcp_config_path and self._mcp is None:
            self._mcp = MCPClient.from_config(self.config.mcp_config_path)
        elif self._mcp is None:            
            self._mcp = MCPClient()
        return self._mcp
    
    @mcp.setter
    def mcp(self, client :MCPClient):
        self._mcp = client

    @staticmethod
    def _to_provider_tool_schema(tool_schema: ToolSchema)->Dict[str, Any]:
        raise NotImplementedError("the selected model-provider does not support tool calling")
    
    @staticmethod
    def _to_provider_tool_call_schema(toolCallSchema :ToolCallSchema)->ToolCallSchema:        
        raise NotImplementedError("the selected model-provider does not support tool calling")

    def _tool_call_message(self, **kwargs)->Dict[str, str]:        
        raise NotImplementedError("the selected model-provider does not support tool calling")

    @staticmethod
    def get_default_tokenizer(model_name: str) -> str:
        """Get default tokenizer name for a model.
        
        Args:
            model_name: Name of the model to get tokenizer for
            
        Returns:
            str: Tokenizer name (falls back to 'gpt-4o' if unknown)
        """
        try:
            tiktoken.encoding_name_for_model(model_name)
            return model_name
        except KeyError:
            return "gpt-4o"

    @staticmethod
    def _message_content(prompt: Union[List[str], str], img_b64_str: Optional[List[str]] = None) -> Union[str, List[Dict]]:
        """Format message content for API requests.
        
        Args:
            prompt: Input prompt(s) to format
            img_b64_str: Optional list of base64 encoded images
            
        Returns:
            Union[str, List[Dict]]: Formatted message content as string or list of content parts
        """
        if isinstance(prompt, str):
            prompt = [prompt]

        if img_b64_str is None:
            message_content = "\n".join(prompt)
        else:
            message_content = [
                {
                    "type": "text",
                    "text": _prompt
                } for _prompt in prompt
            ]
            for img in img_b64_str:
                message_content.append(
                    {
                        "type": "image_url",
                        "image_url": {"url": f"data:image/jpeg;base64,{img}"},
                    }
                )
        return message_content
    
    @staticmethod
    def async_partial(func, *args, **kwargs):
        """Create async partial function with bound arguments.
        
        Args:
            func: Async function to partialize
            *args: Positional arguments to bind
            **kwargs: Keyword arguments to bind
            
        Returns:
            Callable: Async function with bound arguments
            
        Example:
            >>> async def test(a, b): return a + b
            >>> partial = async_partial(test, 1)
            >>> await partial(2)  # Returns 3
        """
        @wraps(func)
        async def wrapped(*inner_args, **inner_kwargs):
            return await func(*args, *inner_args, **kwargs, **inner_kwargs)
        return wrapped
    
    def use_as_reasoner(self, 
                        session_id :Optional[str]=None,
                        workspace :Optional[str]=None,
                        stop_thinking_token: str = REASONING_STOP_TOKEN):
        """Configure provider as a reasoning assistant.
        
        Args:
            session_id: Optional session identifier to set
            workspace: Optional workspace identifier to set
            stop_thinking_token: Token to stop reasoning output
            
        Example:
            >>> provider.use_as_reasoner(
            ...     session_id="123",
            ...     stop_thinking_token="[STOP_REASONING]"
            ... )
        """
        if self.session_id:
            self.session_id = session_id
        self.worspace = workspace
        self.completion_fn = partial(self.completion_fn, stop=stop_thinking_token)
        self.acompletion_fn = self.async_partial(self.acompletion_fn, stop=stop_thinking_token)
        self._is_reasoner = True

    def _message_body(self, prompt: Union[List[str], str], role: Literal["user", "system", "assistant"] = "user", img_b64_str: Optional[List[str]] = None, _last: Optional[bool] = False) -> Dict:
        """Create message body for API requests.
        
        Args:
            prompt: Input prompt(s) to include
            role: Message role (user/system/assistant)
            img_b64_str: Optional base64 encoded images
            _last: Whether this is the last message
            
        Returns:
            Dict: Formatted message body
        """
        message_body = {
            "role": role,
            "content": self._message_content(prompt, img_b64_str)
        }
        return message_body

    @staticmethod
    def _validte_message_dict(message_dict: Dict[str, str]) -> bool:
        """Validate message dictionary structure.
        
        Args:
            message_dict: Message dictionary to validate
            
        Returns:
            bool: True if valid
            
        Raises:
            AssertionError: If message structure is invalid
        """
        assert message_dict.get("role") in ["user", "system", "assistant", "tool"], f"{message_dict} 'role' attribute must be one of ['user', 'system', 'assistant', 'tool]"
        assert message_dict.get("content") is not None or message_dict.get("tool_calls") is not None, f"{message_dict} 'content' or 'tool_calls' attribute is missing"
        return True

    def _map_multiple_prompts(self, prompt: Union[List[str], List[Dict[str, str]]]) -> List[str]:
        """Map multiple prompts to message sequence with alternating roles.
        
        Args:
            prompt: List of prompts or message dictionaries
            
        Returns:
            List[str]: Sequence of formatted messages with alternating roles
        """
        next_role_maps = {
            "assistant": "user",
            "user": "assistant"
        }
        role = "user"
        prompt_messages = []
        for _prompt in prompt[::-1]:
            if isinstance(_prompt, str):
                _prompt = self._message_body(_prompt, role=role)

            elif isinstance(_prompt, ToolCallSchema):
                _prompt = _prompt._raw
            
            elif isinstance(_prompt, dict):
                self._validte_message_dict(_prompt)
                role = _prompt.get("role")
            
            role = next_role_maps.get(role)
            prompt_messages.append(_prompt)

        return prompt_messages[::-1]
    
    def _handle_system_prompt(self,
        messages :list,
        system_prompt: Optional[Union[List[str], str]] = None):
        """Handle system prompt for API requests.
        
        Args:
            messages: List of messages to append to
            system_prompt: System prompt content to add
        """
        if system_prompt is not None:
            messages.append(self._message_body(system_prompt, role="system"))

    def _handle_special_sys_prompt_anthropic(self, args :Dict, system_prompt: Optional[Union[List[str], str]] = None):
        """placeholder to be overwritten by the anthropic provider"""
        pass

    def _handle_openai_response_only_models(self, args :Dict):
        """Placeholder to be overwritten by the openai provider"""
        pass

    @classmethod
    def _handle_tools(cls, tools  :List[ToolSchema])->Dict:
        if not tools:
            return None
        return [
            cls._to_provider_tool_schema(tool)
            for tool in tools
        ]

    def completion_args_template(self,
        prompt: Union[str, List[str], List[Dict[str, str]]],
        system_prompt: Optional[Union[List[str], str]] = None,
        prefix_prompt: Optional[Union[List[str], str]] = None,
        img_b64_str: Optional[Union[str, List[str]]] = None,
        tools: Optional[List[ToolSchema]]=None,
        stream: bool = False) -> Dict:
        """Create completion arguments template for API requests.
        
        Args:
            prompt: Input prompt(s)
            system_prompt: Optional system prompt
            prefix_prompt: Optional prefix prompt
            img_b64_str: Optional base64 encoded images
            stream: Whether to stream response
            
        Returns:
            Dict: Prepared completion arguments
        """
        if img_b64_str and isinstance(img_b64_str, str):
            img_b64_str = [img_b64_str]
        if isinstance(prompt, str):
            prompt = [prompt]

        messages = []
        self._handle_system_prompt(messages, system_prompt)

        messages.extend(self._map_multiple_prompts(prompt))

        if prefix_prompt is not None:
            messages.append(self._message_body(prefix_prompt, role="assistant", _last=True))
        
        args = dict(
            model=self.config.model,
            temperature=self.config.temperature,
            max_tokens=self.config.max_tokens,
            messages=messages,
            tools=tools,
            tool_choice=self.config.tool_choice,
            stream=stream
        )
        
        if self.completion_args:
            if not stream:
                self.completion_args.pop("stream_options", None)
            args.update(self.completion_args)

        self._handle_openai_response_only_models(args)
        self._handle_special_sys_prompt_anthropic(args, system_prompt)
        
        args = {arg: value for arg, value in args.items() if value is not None}
        
        return args
    
    @staticmethod
    def _img_to_base64(img_path :Optional[Union[Union[str, Path, bytes], List[Union[str, Path, bytes]]]] = None):
        return [
            image_to_base64(img)
            for img in img_path
        ] if img_path else None

    def _prepare_completion_args(self,
        prompt: Union[str, List[str], List[Dict[str, str]]], 
        system_prompt: Optional[Union[List[str], str]] = None,
        prefix_prompt: Optional[Union[List[str], str]] = None,
        img_b64_str: Optional[List[str]] = None,
        tools: Optional[List[ToolSchema]]=None,
        stream: bool = True) -> Dict: 
        """Prepare completion arguments including image processing.
        
        Args:
            prompt: Input prompt(s)
            system_prompt: Optional system prompt
            prefix_prompt: Optional prefix prompt
            img_path: Optional image path(s)
            stream: Whether to stream response
            
        Returns:
            Dict: Prepared completion arguments
        """

        completion_args = self.completion_args_template(
            prompt=prompt,
            system_prompt=system_prompt,
            prefix_prompt=prefix_prompt,
            img_b64_str=img_b64_str,
            tools=tools,
            stream=stream
        )
        return completion_args
    
    @staticmethod
    def _handle_reasoning_steps(chunk_message, message, _skip)->bool:
        """Handle reasoning steps in streamed responses.
        
        Args:
            chunk_message: Current message chunk
            message: Accumulated message
            _skip: Whether to skip current chunk
            
        Returns:
            bool: Updated skip state
        """
        if chunk_message == REASONING_START_TOKEN:
            _skip = True
        message.append(chunk_message) if not _skip else ... 
        if chunk_message == REASONING_STOP_TOKEN:
            _skip = False
        return _skip
    
    @classmethod
    def _is_tool_call(cls, _chunk)->bool:
        if _chunk and hasattr(cls._chunk_from_provider(_chunk).delta, "tool_calls") and cls._chunk_from_provider(_chunk).delta.tool_calls:
            return True
        return False

    @classmethod
    def _handle_stream_messages(cls, _chunk, message, _skip=False)->bool:
        """Handle streamed messages from synchronous completions.
        
        Args:
            _chunk: Raw message chunk
            message: Accumulated message
            _skip: Whether to skip current chunk
            
        Returns:
            bool: Updated skip state
        """
        chunk_message = _chunk[0].delta.content or ""
        default_stream_handler(chunk_message)
        return cls._handle_reasoning_steps(chunk_message, message, _skip)
    
    @classmethod
    async def _handle_astream_messages(cls, _chunk, logger_fn, message, _skip=False)->bool:
        """Handle streamed messages from asynchronous completions.
        
        Args:
            _chunk: Raw message chunk
            logger_fn: Async logging function
            message: Accumulated message
            _skip: Whether to skip current chunk
            
        Returns:
            bool: Updated skip state
        """
        chunk_message = _chunk[0].delta.content or  ""
        await logger_fn(chunk_message)
        return cls._handle_reasoning_steps(chunk_message, message, _skip)
    
    @staticmethod
    def _chunk_from_provider(_chunk):
        return _chunk[0]
    
    @classmethod
    def _tool_chunk_from_provider(cls, _chunk):
        return cls._chunk_from_provider(_chunk).delta.tool_calls[0]
    
    @staticmethod
    def _fill_tool_schema(tool_chunk)->ToolCallSchema:
        tool_call = ToolCallSchema(
            id=tool_chunk.id,
            name=tool_chunk.function.name,
            arguments=tool_chunk.function.arguments
        )
        tool_call._raw = tool_chunk.function
        return tool_call
    
    @staticmethod
    def _tool_call_change_condition(tool_chunk)->bool:
        return tool_chunk.id is not None
    
    @staticmethod
    def _handle_tool_call_stream(tool_call :ToolCallSchema, tool_chunk)->ToolCallSchema:
        tool_call._raw.arguments += tool_chunk.function.arguments
        tool_call.arguments += tool_chunk.function.arguments
        return tool_call

    def _no_stream(self, response) -> Union[str, ToolCalls]:
        _chunk = self.normalize_fn(response)
        message = self._chunk_from_provider(_chunk).message
        if hasattr(message, "tool_calls") and message.tool_calls:
            return ToolCalls(root=[
                self._fill_tool_schema(tool_call) for tool_call in message.tool_calls
            ])
        else:
            return message.content

    def _stream(self, stream, prefix_prompt: Optional[Union[str, List[str]]] = None) -> str:
        """Handle streaming response from synchronous completion.
        
        Args:
            stream: Response stream
            prefix_prompt: Optional prefix prompt
            
        Returns:
            str: Accumulated response
        """
        message = [] 
        _skip = False
        completeion_id = ulid.ulid()
        for chunk in stream:
            _chunk = self.normalize_fn(chunk, completeion_id)
            if _chunk:
                _skip = self._handle_stream_messages(_chunk, message, _skip)

        if self._is_reasoner:
            default_stream_handler(REASONING_STOP_TOKEN)
        else:
            default_stream_handler(STREAM_END_TOKEN)

        response = "".join(message)
        return response
    
    async def _astream(self, stream, logger_fn, prefix_prompt: Optional[Union[str, List[str]]] = None) -> Union[str, ToolCalls]:
        """Handle streaming response from asynchronous completion.
        
        Args:
            stream: Async response stream
            logger_fn: Async logging function
            prefix_prompt: Optional prefix prompt
            
        Returns:
            str: Accumulated response
        """
        message :List[Union[str, ToolCallSchema]] = []
        _skip = False
        completeion_id = ulid.ulid()
        await logger_fn(STREAM_START_TOKEN) if not prefix_prompt else ...
        _calling_tool = False
        tool_calls = ToolCalls()
        async for chunk in stream:
            _chunk = self.normalize_fn(chunk, completeion_id)
            if _chunk:
                _skip = await self._handle_astream_messages(_chunk, logger_fn, message, _skip)

            if self._is_tool_call(_chunk):
                ### TODO recheck this line to ensure it covers multiple tool calling in stream mode
                tool_chunk = self._tool_chunk_from_provider(_chunk)
                if not _calling_tool:
                    _calling_tool = True
                    await logger_fn(TOOL_CALL_START_TOKEN)
                    tool_call = self._fill_tool_schema(tool_chunk)
                    continue

                if self._tool_call_change_condition(tool_chunk):
                    if message:
                        ### cover anthropic
                        tool_call._raw = "\n".join(message)
                    tool_calls.root.append(tool_call)
                    tool_call = self._fill_tool_schema(tool_chunk)
                    continue
                
                tool_call = self._handle_tool_call_stream(tool_call, tool_chunk)
        
        if _calling_tool:
            ### colect last call
            if message:
                ### cover anthropic
                tool_call._raw = "\n".join(message)
            tool_calls.root.append(tool_call)
            return tool_calls
        
        if self._is_reasoner:
            await logger_fn(REASONING_STOP_TOKEN)
        else:
            await logger_fn(STREAM_END_TOKEN)
        response = "".join(message)
        return response
    
    async def connect_to_mcp(self):
        if self.mcp is not None and self.config.tool_use is not False:
            if not self.mcp._is_connected or self.mcp._needs_update:
                await self.mcp.connect()
                tools = await self.mcp.servers.tools
                self.tools = self._handle_tools(tools)
                self.mcp._needs_update = False

    @property
    def _has_not_exceeded_tool_calls(self)->bool:
        return self.config.max_tool_calls_per_response is None  or self._n_sucessive_tool_calls < self.config.max_tool_calls_per_response
    
    async def _execute_tool_calls(self, output :list)->Tuple[list, bool]:
        """
        Execute multiple tool calls in parallel using asyncio.gather.
        
        Args:
            output: The output containing potential tool calls
            logger: Logger instance for logging execution information
        
        Returns:
            List of tool messages generated from tool calls
        """
        tools_messages = []
        call_tool = False
        
        for _output in output:
            if isinstance(_output, ToolCalls) and self._has_not_exceeded_tool_calls:
                call_tool = True
                tool_calls = _output.root
                
                # If there are multiple tool calls, execute them in parallel
                if len(tool_calls) > 1 and self.config.concurrent_tool_calls:
                    
                    tool_names = [tool_call.name for tool_call in tool_calls]
                    # Log the start of tool execution
                    _logger.logger.info(f"Executing {len(tool_calls)} tools: {tool_names}")
                    start_time = time.time()
                    
                    tool_coroutines = [
                        self.mcp.servers.call_tool(
                            tool_name=tool_call.name,
                            arguments=json.loads(repair_json(tool_call.arguments) or "{}"),
                            silent=True
                        )
                        for tool_call in tool_calls
                    ]
                    
                    # Execute all tool calls in parallel
                    tool_responses = await asyncio.gather(*tool_coroutines)
                    
                    # Process the responses
                    for tool_call, tool_call_response in zip(tool_calls, tool_responses):
                        tools_messages.extend([
                            self._to_provider_tool_call_schema(tool_call),
                            self._tool_call_message(toolCallSchema=tool_call, content=tool_call_response),
                        ])

                    execution_time = time.time() - start_time
                    _logger.logger.info(f"Finished executing {len(tool_calls)} tools in {execution_time:.2f} seconds")
                else:
                    # If there's only one tool call, execute it directly
                    for tool_call in tool_calls:
                        tool_call_response = await self.mcp.servers.call_tool(
                            tool_name=tool_call.name,
                            arguments=json.loads(repair_json(tool_call.arguments) or "{}"),
                        )
                        tools_messages.extend([
                            self._to_provider_tool_call_schema(tool_call),
                            self._tool_call_message(toolCallSchema=tool_call, content=tool_call_response)
                        ])
                    
            elif isinstance(_output, str):
                tools_messages.append(self._message_body(_output, role="assistant"))
        
        return tools_messages, call_tool

    @staticmethod
    def model_to_str(model: Union[BaseModel, RootModel]) -> str:
        """Convert model to JSON string representation."""
        return f"```json\n{model.model_dump_json(indent=4)}\n```"
    
    @staticmethod
    def extract_json(output: str) -> Dict:
        try:
            return json.loads(parse_content(output))
        except json.JSONDecodeError:
            return output
  
    @staticmethod
    def _get_message_records(completion_args :Dict[str, str], excluded_roles :Optional[List[Literal["user", "system", "assistant"]]]=None)->List[Dict[str, str]]:
        messages = completion_args.get("messages")
        records = [
            message for message in messages
            if not excluded_roles or message.get("role") not in excluded_roles
        ]
        return records

    def complete(
        self,
        prompt: Union[str, List[str], List[Dict[str, str]], BaseModel, RootModel], 
        system_prompt: Optional[Union[str, List[str]]] = None,
        prefix_prompt: Optional[Union[str, List[str]]] = None,
        img_path: Optional[Union[Union[str, Path, bytes], List[Union[str, Path, bytes]]]] = None,
        json_output: bool = False,
        stream: bool = True,
        agent_id: Optional[str]=None,
        action_id :Optional[str]=None) -> Union[str, Dict]:
        """
        Complete a prompt using the LLM provider.
            
        Args:
            prompt: Input prompt (can be str, BaseModel, or RootModel)
            system_prompt: Optional system prompt override
            prefix_prompt: Additional context to prepend
            img_path: Optional image path(s) for multimodal input
            json_output: Whether to parse output as JSON
            stream: Whether to stream the response
            agent_id: Optional agent identifier
            action_id: Optional action identifier
            
        Returns:
            The completion result as either a string or dictionary (if json_output=True)
        """
        img_b64_str = self._img_to_base64(img_path)
        if isinstance(prompt, Union[BaseModel, RootModel]):
            prompt = self.model_to_str(prompt)
        
        # Start tracking operation time
        start_time = time.time()
        input_tokens = 0
        output_tokens = 0
        cost = 0
        
        completion_args = self._prepare_completion_args(
            prompt=prompt,
            system_prompt=system_prompt,
            prefix_prompt=prefix_prompt,
            img_b64_str=img_b64_str,
            stream=stream
        )
        
        output = None
        error_message = None
        try:
            output = self.completion_fn(**completion_args)
            
            if stream:
                output = self._stream(output, prefix_prompt)
            else:
                output = self._no_stream(output)

            if self.usage:
                if self.usage.latest_completion:
                    _logger.logger.info(str(self.usage.latest_completion))
                    input_tokens = self.usage.latest_completion.prompt_tokens
                    output_tokens = self.usage.latest_completion.response_tokens
                    cost = self.usage.latest_completion.cost
                _logger.logger.info(str(self.usage))
            
            output = output if not json_output else self.extract_json(output)

        except Exception as e:
            error_message = str(e)
            output = None 
            raise e
        
        finally:
            end_time = time.time()
            latency_ms = (end_time - start_time) * 1000
            
            if self.collector:
                self.collector.record_completion(
                    provider=self.config.provider,
                    operation_type="completion",
                    completion_args=completion_args,
                    response=output,
                    session_id=self.session_id,
                    workspace=self.worspace,
                    agent_id=agent_id or self.agent_id,
                    action_id=action_id,
                    input_tokens=input_tokens,
                    output_tokens=output_tokens,
                    cost=cost,
                    latency_ms=latency_ms,
                    error_message=error_message,
                    extras=self.extras
                )
            
        return output

    async def acomplete(
        self,
        prompt: Union[str, List[str], List[Dict[str, str]], BaseModel, RootModel],
        system_prompt: Optional[Union[str, List[str]]] = None,
        prefix_prompt: Optional[Union[str, List[str]]] = None,
        img_path: Optional[Union[Union[str, Path], List[Union[str, Path]]]] = None,
        json_output: bool = False,
        stream: bool = True,       
        as_message_records :bool=False,
        stream_handler: Optional[Callable[[str], None]] = default_stream_handler,
        agent_id: Optional[str]=None,
        action_id :Optional[str]=None
        ) -> Union[str, Dict, List[Union[str, Dict[str, str]]]]:
        """
        Async version of complete() to generate completions.
        
        Args:
            prompt: Input prompt (can be str, list, dict, BaseModel or RootModel)
            system_prompt: Optional system prompt override
            prefix_prompt: Additional context to prepend
            img_path: Optional image path(s) for multimodal input
            json_output: Whether to parse output as JSON
            stream: Whether to stream the response
            stream_handler: default_stream_handler
            agent_id: Optional agent identifier
            action_id: Optional action identifier
            
        Returns:
            The completion result as either a string or dictionary (if json_output=True)
        """
        img_b64_str = self._img_to_base64(img_path)
        if isinstance(prompt, Union[BaseModel, RootModel]):
            prompt = self.model_to_str(prompt)
        elif isinstance(prompt, str):            
            prompt = [self._message_body([prompt], role="user", img_b64_str=img_b64_str)]
        
        stream_handler = stream_handler or default_stream_handler
        
        # Start tracking operation time
        start_time = time.time()
        input_tokens = 0
        output_tokens = 0
        cost = 0

        await self.connect_to_mcp()
        
        completion_args = self._prepare_completion_args(
            prompt=prompt,
            system_prompt=system_prompt,
            prefix_prompt=prefix_prompt,
            img_b64_str=img_b64_str,
            tools=self.tools if self._has_not_exceeded_tool_calls else None,
            stream=stream
        )
        output = None 
        call_tool = False
        error_message = None
        try:
            output = await self.acompletion_fn(**completion_args)
            
            if stream:
                output = await self._astream(output, stream_handler, prefix_prompt)
            else:
                output = self._no_stream(output)
            
            if self.usage:
                if self.usage.latest_completion:
                    _logger.logger.info(str(self.usage.latest_completion))
                    input_tokens = self.usage.latest_completion.prompt_tokens
                    output_tokens = self.usage.latest_completion.response_tokens
                    cost = self.usage.latest_completion.cost
                _logger.logger.info(str(self.usage))
            
            ### handle scenarios of text + toolcalssblock i.e anthropic
            _is_not_list = False
            if not isinstance(output, list):
                _is_not_list = True
                output = [output]
            
            tools_messages, call_tool = await self._execute_tool_calls(output)

            if call_tool:
                output = [
                    _.model_dump() if isinstance(_, ToolCalls) else _ for _ in output
                ]
                if stream:
                    await stream_handler(TOOL_CALL_END_TOKEN)
                
                messages = completion_args.get("messages")
                messages = [_ for _ in messages]
                messages.extend(tools_messages)
                self._n_sucessive_tool_calls += 1

                if self.collector:
                    end_time = time.time()
                    latency_ms = (end_time - start_time) * 1000
                    asyncio.create_task(
                        self.collector.arecord_completion(
                            provider=self.config.provider,
                            operation_type="acompletion.tool_call",
                            completion_args=completion_args,
                            response=output.model_dump_json(indent=4) if isinstance(output, ToolCalls) else output,
                            session_id=self.session_id,
                            workspace=self.worspace,
                            agent_id=agent_id or self.agent_id,
                            action_id=action_id,
                            input_tokens=input_tokens,
                            output_tokens=output_tokens,
                            cost=cost,
                            latency_ms=latency_ms,
                            error_message=error_message,
                            extras=self.extras
                        ), 
                        name="collector.arecord_completion.tool_call"
                    )

                return await self.acomplete(
                    prompt=messages,
                    system_prompt=system_prompt,
                    prefix_prompt=prefix_prompt,
                    img_path=img_path,
                    json_output=json_output,
                    stream=stream,
                    as_message_records=as_message_records,
                    stream_handler=stream_handler,
                    agent_id=agent_id,
                    action_id=action_id
                )
            
            self._n_sucessive_tool_calls = 0
            if _is_not_list:
                output = output[-1]
            
            output = output if not json_output else self.extract_json(output)

        except Exception as e:
            error_message = str(e)
            output = None 
            raise e
        
        finally:
            if not call_tool:
                if self.collector:
                    end_time = time.time()
                    latency_ms = (end_time - start_time) * 1000
                    asyncio.create_task(
                        self.collector.arecord_completion(
                            provider=self.config.provider,
                            operation_type="acompletion",
                            completion_args=completion_args,
                            response=output,
                            session_id=self.session_id,
                            workspace=self.worspace,
                            agent_id=agent_id or self.agent_id,
                            action_id=action_id,
                            input_tokens=input_tokens,
                            output_tokens=output_tokens,
                            cost=cost,
                            latency_ms=latency_ms,
                            error_message=error_message,
                            extras=self.extras
                        ),
                        name="collector.arecord_completion"
                    )
        
        if as_message_records:
            records = self._get_message_records(completion_args, excluded_roles=["system"])
            records.append(output)
            output = records
            
        return output


from typing import Literal, Optional, Union, Dict
from typing_extensions import Self
from pydantic import BaseModel, field_validator, model_validator, ConfigDict

from aicore.const import DEFAULT_TIMEOUT, SUPPORTED_REASONER_PROVIDERS, SUPPORTED_REASONER_MODELS
from aicore.models_metadata import METADATA, PricingConfig

class LlmConfig(BaseModel):
    provider :Literal["anthropic", "gemini", "groq", "mistral", "nvidia", "openai", "openrouter", "deepseek", "grok"]
    api_key :str
    model :str
    base_url :Optional[str]=None
    temperature :float=0
    max_tokens :int=12000
    reasoner :Optional["LlmConfig"]=None
    pricing :Optional[PricingConfig]=None
    _context_window :Optional[int]=None

    mcp_config_path :Optional[str]=None
    tool_choice :Union[str, Dict, None]=None
    max_tool_calls_per_response :Optional[int]=None
    concurrent_tool_calls :Optional[bool]=True

    timeout :Optional[int]=DEFAULT_TIMEOUT
    tool_use :Optional[bool]=None

    model_config = ConfigDict(
        extra="allow",
    )

    @field_validator("temperature")
    @classmethod
    def ensure_temperature_lower_than_unit(cls, temperature :float)->float:
        assert 0 <= temperature <= 1, "temperature should be between 0 and 1"
        return temperature
    
    @field_validator("reasoner", mode="after")
    @classmethod
    def ensure_valid_reasoner(cls, reasoner :"LlmConfig")->"LlmConfig":
        if isinstance(reasoner, LlmConfig):
            assert reasoner.provider in SUPPORTED_REASONER_PROVIDERS, f"{reasoner.provider} is not supported as a reasoner provider. Supported providers are {SUPPORTED_REASONER_PROVIDERS}"
            assert reasoner.model in SUPPORTED_REASONER_MODELS, f"{reasoner.model} is not supported as a reasoner model. Supported models are {SUPPORTED_REASONER_MODELS}"
        return reasoner
    
    @property
    def provider_model(self)->str:
        return f"{self.provider}-{self.model}"
    
    @property
    def context_window(self)->int:
        return self._context_window
    
    @context_window.setter
    def context_window(self, value :int):
        self._context_window = value

    @model_validator(mode="after")
    def initialize_pricing_from_defaults(self)->Self:
        model_metadata =  METADATA.get(self.provider_model)
        if model_metadata is not None:
            if self.pricing is None and model_metadata.pricing is not None:
                self.pricing = model_metadata.pricing
            if self.max_tokens > model_metadata.max_tokens:
                self.max_tokens = model_metadata.max_tokens
            if self.context_window is None and model_metadata.context_window:
                self.context_window = model_metadata.context_window
            if self.tool_use is None:
                self.tool_use = model_metadata.tool_use
        
        return self
    
    @field_validator("tool_choice", mode="before")
    @classmethod
    def auto_tool_choice_from_mcp(cls, kwargs :Dict)->Dict:
        if kwargs and kwargs.get("mcp_config_path") and not kwargs.get("tool_choice"):
            kwargs["tool_choice"] = "auto"
        return kwargs


from aicore.llm.providers.base_provider import LlmBaseProvider
from aicore.llm.mcp.models import ToolCallSchema, ToolSchema
from aicore.const import OPENAI_RESPONSE_API_MODELS, OPENAI_RESPONSE_ONLY_MODELS
from pydantic import model_validator
from openai import OpenAI, AsyncOpenAI, AuthenticationError
from openai.types.chat import ChatCompletion
from typing import Any, Dict, Optional
from typing_extensions import Self
import tiktoken

class OpenAiLlm(LlmBaseProvider):
    base_url :Optional[str]=None

    @model_validator(mode="after")
    def set_openai(self)->Self:

        self.client :OpenAI = OpenAI(
            api_key=self.config.api_key,
            base_url=self.base_url or self.config.base_url
        )
        _aclient :AsyncOpenAI = AsyncOpenAI(
            api_key=self.config.api_key,
            base_url=self.base_url or self.config.base_url
        )
        self._auth_exception = AuthenticationError
        self.validate_config()
        self.aclient = _aclient
        
        if self.config.model in OPENAI_RESPONSE_ONLY_MODELS:
            self.completion_fn = self.client.responses.create
            self.acompletion_fn = _aclient.responses.create
        else:
            self.completion_fn = self.client.chat.completions.create
            self.acompletion_fn = _aclient.chat.completions.create
            
        self.completion_args["stream_options"] = {
            "include_usage": True
        }
        self.normalize_fn = self.normalize

        self.tokenizer_fn = tiktoken.encoding_for_model(
            self.get_default_tokenizer(
                self.config.model
            )
        ).encode

        self._handle_reasoning_models()

        return self
    
    def normalize(self, chunk :ChatCompletion, completion_id :Optional[str]=None):
        usage = chunk.usage
        if usage is not None:
            cached_tokens = usage.prompt_tokens_details.cached_tokens \
            if usage.prompt_tokens_details is not None \
            else 0
            ### https://platform.openai.com/docs/guides/prompt-caching
            self.usage.record_completion(
                prompt_tokens=usage.prompt_tokens-cached_tokens,
                response_tokens=usage.completion_tokens,
                cached_tokens=cached_tokens,
                completion_id=completion_id or chunk.id
            )
        return chunk.choices
    
    def _handle_reasoning_models(self):
        ### o series models
        if self.config.model.startswith("o") or self.config.model in OPENAI_RESPONSE_API_MODELS:
            self.completion_args["temperature"] = None
            self.completion_args["max_tokens"] = None
            self.completion_args["max_completion_tokens"] = self.config.max_tokens
            reasoning_efftort = getattr(self.config, "reasoning_efftort", None)
            if reasoning_efftort is not None:
                self.completion_args["reasoning_efftort"] = reasoning_efftort

    def _handle_openai_response_only_models(self, args :Dict):
        if self.config.model in OPENAI_RESPONSE_ONLY_MODELS:
            args["input"] = args.pop("messages")
            args.pop("stream_options")
            args.pop("max_completion_tokens")

    @staticmethod
    def _to_provider_tool_schema(tool: ToolSchema) -> Dict[str, Any]:
        """
        Convert to OpenAi tool schema format.
        
        Returns:
            Dictionary in OpenAi tool schema format
        """
        return {
            "type": "function",
            "function": {
                "name": tool.name,
                "description": tool.description,
                "parameters": {
                    "type": tool.input_schema.type,
                    "properties": tool.input_schema.properties.model_dump(),
                    "required": tool.input_schema.required,
                    **{k: v for k, v in tool.input_schema.model_dump().items() 
                       if k not in ["type", "properties", "required"]}
                }
            }
        }
    
    @staticmethod
    def _to_provider_tool_call_schema(toolCallSchema :ToolCallSchema)->ToolCallSchema:
        toolCallSchema._raw = {
            "role": "assistant",
            "tool_calls": [
                {
                    "id": toolCallSchema.id,
                    "function": {
                        "name": toolCallSchema.name,
                        "arguments": toolCallSchema.arguments
                    },
                    "type": "function"
                }
            ]
        }
        
        # ChatCompletionMessage(
        #     role="assistant",
        #     tool_calls=[
        #         ChatCompletionMessageToolCall(
        #             id=toolCallSchema.id,
        #             function=Function(
        #                 name=toolCallSchema.name,
        #                 arguments=toolCallSchema.arguments
        #             ),
        #             type="function"
        #         )
        #     ]
        # )

        return toolCallSchema
    
    def _tool_call_message(self, toolCallSchema :ToolCallSchema, content :str) -> Dict[str, str]:
        return {
            "type": "function_call_output",
            "role": "tool",
            "tool_call_id": toolCallSchema.id,
            "content": str(content)
        }


import pytest
from unittest.mock import MagicMock, AsyncMock, patch
from aicore.llm.llm import Llm
from aicore.llm.config import LlmConfig
import json

class MockResponse:
    def __init__(self, chunks, usage=None):
        self.chunks = [chunks]
        self.usage = usage

    def __iter__(self):
        for chunks in self.chunks:
            yield chunks

    async def __aiter__(self):
        for chunks in self.chunks:
            yield chunks

class MockMistralResponse:
    def __init__(self, chunks):
        self.chunks = [(chunks)]

    def __iter__(self):
        for chunks in self.chunks:
            yield chunks

    async def __aiter__(self):
        for chunks in self.chunks:
            yield chunks
class MockChunk:
    def __init__(self, chunk_content):
        self.choices = [MockChoices(chunk_content)]
        self.usage = None
        self.x_groq = None

class MockMistralChunk:
    def __init__(self, data):
        self.data = MockData(data)

class MockData:
    def __init__(self, choices):
        self.choices = [MockChoices(choices)]       
        self.usage = None
class MockChoices:
    def __init__(self, delta):
        self.delta = MockDelta(delta)

class MockDelta:
    def __init__(self, content):
        self.content = content

class MockUsage:
    def __init__(self, total_tokens):
        self.total_tokens = total_tokens

def MockGeminiTokenizer(content):
    return [_ for _ in content.split(" ")]

@pytest.fixture
def mock_openai(*args, **kwargs):
    mock_openai = MagicMock()
    mock_openai.complete = MagicMock(return_value=MockResponse(MockChunk("mocked response")))
    mock_openai.acomplete = AsyncMock(return_value=MockResponse(MockChunk("mocked async response")))
    return mock_openai

@pytest.fixture
def mock_openai_json(*args, **kwargs):
    mock_openai = MagicMock()
    mock_openai.complete = MagicMock(return_value=MockResponse(MockChunk(json.dumps({"response": "mocked"}))))
    mock_openai.acomplete = AsyncMock(return_value=MockResponse(MockChunk(json.dumps({"response": "async mocked"}))))
    return mock_openai

@pytest.fixture
def mock_mistral(*args, **kwargs):
    mock_mistral = MagicMock()
    mock_mistral.complete = MagicMock(return_value=MockMistralResponse(MockMistralChunk("mocked response")))
    mock_mistral.acomplete = AsyncMock(return_value=MockMistralResponse(MockMistralChunk("mocked async response")))
    return mock_mistral

@pytest.fixture
def mock_mistral_json(*args, **kwargs):
    mock_mistral = MagicMock()
    mock_mistral.complete = MagicMock(return_value=MockMistralResponse(MockMistralChunk(json.dumps({"response": "mocked"}))))
    mock_mistral.acomplete = AsyncMock(return_value=MockMistralResponse(MockMistralChunk(json.dumps({"response": "async mocked"}))))
    return mock_mistral

@pytest.fixture
def mock_gemini_tokenizer(*args, **kwargs):
    mock_gemini = MagicMock()
    mock_gemini.tokenizer_fn = MagicMock(return_value=MockGeminiTokenizer("Hi there"))
    return mock_gemini

@pytest.fixture
def llm_config_openai():
    return LlmConfig(provider="openai", api_key="test_key", model="gpt-4")

@pytest.fixture
def llm_config_mistral():
    return LlmConfig(provider="mistral", api_key="test_key", model="mistral-7b")

@pytest.fixture
def llm_config_groq():
    return LlmConfig(provider="groq", api_key="test_key", model="mixtral-8x7b-32768")

@pytest.fixture
def llm_config_gemini():
    return LlmConfig(provider="gemini", api_key="test_key", model="gemini-pro")

@pytest.fixture
def llm_config_nvidia():
    return LlmConfig(provider="nvidia", api_key="test_key", model="nemotron-4")

@patch('aicore.llm.providers.base_provider.LlmBaseProvider.validate_config')
@pytest.mark.parametrize("provider_name", ["openai", "groq", "gemini", "mistral", "nvidia"])
@pytest.mark.asyncio
async def test_llm_complete(mock_validate_config, provider_name,  mock_openai, mock_mistral, llm_config_openai,
        llm_config_mistral, llm_config_groq, llm_config_gemini, llm_config_nvidia):
    
    config = {
        "openai": llm_config_openai,
        "mistral": llm_config_mistral,
        "groq": llm_config_groq,
        "gemini": llm_config_gemini,
        "nvidia": llm_config_nvidia
    }[provider_name]

    if provider_name == "mistral":
        mock_openai = mock_mistral
    
    mock_validate_config.return_value = None
    # Initialize Llm with the given configuration
    llm = Llm.from_config(config)

    llm.provider.completion_fn = mock_openai.complete
    llm.provider.acompletion_fn = mock_openai.acomplete
    
    # Call the complete method with a test prompt
    prompt = "test prompt"
    # Test with string input
    response = llm.complete(prompt)
    aresponse = await llm.acomplete(prompt)
    # Assert that the response is a string
    assert isinstance(response, str)
    # Assert that the response contains the mocked response
    assert "mocked" in response
    # Assert that the response is a string
    assert isinstance(aresponse, str)
    # Assert that the response contains the mocked response
    assert "mocked" in aresponse
    # Assert that the API was called once
    if provider_name == "openai":
        mock_openai.complete.assert_called_once()
        mock_openai.acomplete.assert_called_once()

@patch('aicore.llm.providers.base_provider.LlmBaseProvider.validate_config')
@pytest.mark.parametrize("provider_name", ["openai", "groq", "gemini", "mistral", "nvidia"])
@pytest.mark.asyncio
async def test_llm_complete_json(mock_validate_config, provider_name, mock_openai_json, mock_mistral_json, llm_config_openai,
        llm_config_mistral, llm_config_groq, llm_config_gemini, llm_config_nvidia):
    
    config = {
        "openai": llm_config_openai,
        "mistral": llm_config_mistral,
        "groq": llm_config_groq,
        "gemini": llm_config_gemini,
        "nvidia": llm_config_nvidia
    }[provider_name]

    if provider_name == "mistral":
        mock_openai_json = mock_mistral_json
    
    mock_validate_config.return_value = None
    # Initialize Llm with the given configuration
    llm = Llm.from_config(config)

    llm.provider.completion_fn = mock_openai_json.complete
    llm.provider.acompletion_fn = mock_openai_json.acomplete
    
    # Call the complete method with a test prompt
    prompt = "test prompt"
    # Test with string input
    response = llm.complete(prompt, json_output=True)
    aresponse = await llm.acomplete(prompt, json_output=True)
    # Assert that the response is a string
    assert isinstance(response, dict)
    # Assert that the response contains the mocked response
    assert "mocked" in response.values()
    # Assert that the response is a string
    assert isinstance(aresponse, dict)
    # Assert that the response contains the mocked response
    assert "async mocked" in aresponse.values()

    # Assert that the API was called once
    if provider_name == "openai":
        mock_openai_json.complete.assert_called_once()
        mock_openai_json.acomplete.assert_called_once()

@patch('aicore.llm.providers.base_provider.LlmBaseProvider.validate_config')
@pytest.mark.parametrize("provider_name", ["openai", "mistral", "gemini", "groq", "nvidia"])
@pytest.mark.asyncio
async def test_llm_tokenizer(mock_validate_config, provider_name, llm_config_openai, llm_config_mistral, llm_config_groq, llm_config_gemini, llm_config_nvidia, mock_gemini_tokenizer):
    config = {
        "openai": llm_config_openai,
        "mistral": llm_config_mistral,
        "groq": llm_config_groq,
        "gemini": llm_config_gemini,
        "nvidia": llm_config_nvidia
    }[provider_name]    
    
    mock_validate_config.return_value = None
    llm = Llm.from_config(config)

    if provider_name == "gemini":
        llm.provider.tokenizer_fn = mock_gemini_tokenizer.tokenizer_fn
    
    prompt = "test prompt"
    
    # Call the tokenizer function
    tokens = llm.tokenizer(prompt)
    
    # Assert that the tokenizer returns a list
    assert isinstance(tokens, list)
    # Assert that the list of tokens is not empty
    assert len(tokens) > 0


DEFAULT_SYSTEM_PROMPT = "You are a helpfull assistant"

REASONER_DEFAULT_SYSTEM_PROMPT = "You are a helpfull assistant with reasoning capabilites that breaks down problems into the detailed steps required to solve them"

REASONING_INJECTION_TEMPLATE = """
Consider the following reasoning steps to help you generate the answer:

{reasoning}
{reasoning_stop_token}
"""


import pytest
import asyncio
from datetime import datetime
from time import time
from typing import AsyncGenerator

# Import the classes from your module
from aicore.logger import Logger, LogEntry

@pytest.fixture
def logger():
    """Fixture to create and cleanup a logger instance for each test."""
    test_logs_dir = "test_logs"
    logger_instance = Logger(logs_dir=test_logs_dir)
    return logger_instance

@pytest.mark.asyncio
async def test_log_chunk_to_queue(logger):
    """Test basic logging functionality with timeout."""
    session_id = "test_session"
    test_message = "Test message"
    
    await logger.log_chunk_to_queue(test_message, session_id)
    
    # Get logs with timeout
    async def get_logs():
        return logger.get_all_logs_in_queue()
    
    logs = await asyncio.wait_for(asyncio.create_task(get_logs()), timeout=2.0)
    
    assert len(logs) == 1
    assert logs[0].message == test_message
    assert logs[0].session_id == session_id
    assert isinstance(logs[0].timestamp, str)

@pytest.mark.asyncio
async def test_pop_with_timeout(logger):
    """Test pop functionality with timeout."""
    session_id = "test_session"
    test_messages = ["Message 1", "Message 2", "Message 3"]
    
    # Log multiple messages
    for msg in test_messages:
        await logger.log_chunk_to_queue(msg, session_id)
    
    received_messages = []
    
    # Create async generator for pop method
    async def collect_messages():
        async for message in logger.pop(session_id):
            received_messages.append(message)
            if len(received_messages) == len(test_messages):
                break
    
    # Run with timeout
    try:
        await asyncio.wait_for(collect_messages(), timeout=2.0)
    except asyncio.TimeoutError:
        pytest.fail("Pop operation timed out")
    
    assert received_messages == test_messages

@pytest.mark.asyncio
async def test_multiple_sessions(logger):
    """Test handling of multiple sessions with timeout."""
    session_1 = "session_1"
    session_2 = "session_2"
    
    # Log messages for different sessions
    await logger.log_chunk_to_queue("Message 1 for session 1", session_1)
    await logger.log_chunk_to_queue("Message 1 for session 2", session_2)
    await logger.log_chunk_to_queue("Message 2 for session 1", session_1)
    
    # Test session queues are created
    assert set(logger.all_sessions_in_queue) == {session_1, session_2}

@pytest.mark.asyncio
async def test_reasoning_stop_token(logger):
    """Test that pop stops when encountering REASONING_STOP_TOKEN."""
    from aicore.logger import REASONING_STOP_TOKEN
    
    session_id = "test_session"
    messages = [
        "Message 1",
        f"Message 2 {REASONING_STOP_TOKEN}",
        "Message 3"  # This shouldn't be received
    ]
    
    for msg in messages:
        await logger.log_chunk_to_queue(msg, session_id)
    
    received_messages = []
    async for message in logger.pop(session_id):
        received_messages.append(message)
    
    assert len(received_messages) == 2
    assert REASONING_STOP_TOKEN in received_messages[1]
    assert messages[2] not in received_messages